# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Based on ipa-server-install code
#
# Copyright (C) 2017-2022  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.0',
    'supported_by': 'community',
    'status': ['preview'],
}

DOCUMENTATION = '''
---
module: ipaserver_enable_ipa
short_description: Enable IPA
description: Enable IPA
options:
  hostname:
    description: Fully qualified name of this host
    type: str
    required: no
  setup_dns:
    description: Configure bind with our zone
    type: bool
    required: yes
  setup_ca:
    description: Configure a dogtag CA
    type: bool
    required: yes
author:
    - Thomas Woerner (@t-woerner)
'''

EXAMPLES = '''
'''

RETURN = '''
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.ansible_ipa_server import (
    check_imports,
    AnsibleModuleLog, setup_logging, options, paths, api, sysrestore, tasks,
    service, bindinstance, redirect_stdout, services
)


def main():
    ansible_module = AnsibleModule(
        argument_spec=dict(
            hostname=dict(required=False, type='str'),
            setup_dns=dict(required=True, type='bool'),
            setup_ca=dict(required=True, type='bool'),
        ),
    )

    ansible_module._ansible_debug = True
    check_imports(ansible_module)
    setup_logging()
    ansible_log = AnsibleModuleLog(ansible_module)

    # set values #############################################################

    options.host_name = ansible_module.params.get('hostname')
    options.setup_dns = ansible_module.params.get('setup_dns')
    options.setup_ca = ansible_module.params.get('setup_ca')

    # Configuration for ipalib, we will bootstrap and finalize later, after
    # we are sure we have the configuration file ready.
    cfg = dict(
        context='installer',
        confdir=paths.ETC_IPA,
        in_server=True,
        # make sure host name specified by user is used instead of default
        host=options.host_name,
    )
    if options.setup_ca:
        # we have an IPA-integrated CA
        cfg['ca_host'] = options.host_name

    api.bootstrap(**cfg)
    api.finalize()
    api.Backend.ldap2.connect()

    # setup ds ######################################################

    fstore = sysrestore.FileStore(paths.SYSRESTORE)

    if hasattr(tasks, "configure_tmpfiles"):
        # Make sure the files we crated in /var/run are recreated at startup
        tasks.configure_tmpfiles()

    if hasattr(service, "enable_services"):
        # Enable configured services and update DNS SRV records
        service.enable_services(options.host_name)
        api.Command.dns_update_system_records()

        if not options.setup_dns:
            # After DNS and AD trust are configured and services are
            # enabled, create a dummy instance to dump DNS configuration.
            bind = bindinstance.BindInstance(fstore)
            bind.create_file_with_system_records()

    with redirect_stdout(ansible_log):
        services.knownservices.ipa.enable()

    ansible_module.exit_json(changed=True)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Copyright (C) 2022  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.0',
    'supported_by': 'community',
    'status': ['preview'],
}

DOCUMENTATION = '''
---
module: ipasmartcard_server_get_vars
short_description:
  Get variables from ipaplatform and ipaserver and python interpreter.
description:
  Get variables from ipaplatform and ipaserver and python interpreter.
author:
    - Thomas Woerner (@t-woerner)
'''

EXAMPLES = '''
- name: Get VARS from IPA
  ipasmartcard_server_get_vars:
  register: ipasmartcard_server_vars
'''

RETURN = '''
NSS_OCSP_ENABLED:
  description:
    Empty string for newer systems using ssl.conf and not nss.conf for
    HTTP instance where OCSP_ENABLED and OCSP_DIRECTIVE are defined in
    ipaserver.install.httpinstance, else NSS_OCSP_ENABLED imported from
    ipaserver.install.httpinstance.
  returned: always
  type: str
NSS_OCSP_DIRECTIVE:
  description:
    Empty string for newer systems using ssl.conf and not nss.conf for
    HTTP instance where OCSP_ENABLED and OCSP_DIRECTIVE are defined in
    ipaserver.install.httpinstance, else NSSOCSP.
  returned: always
  type: str
NSS_NICKNAME_DIRECTIVE:
  description:
    Empty string for newer systems using ssl.conf and not nss.conf for
    HTTP instance where OCSP_ENABLED and OCSP_DIRECTIVE are defined in
    ipaserver.install.httpinstance, else NSSNickname
  returned: always
  type: str
OCSP_ENABLED:
  description:
    OCSP_ENABLED imported from ipaserver.install.httpinstance, if import
    succeeds, else ""
  returned: always
  type: str
OCSP_DIRECTIVE:
  description:
    OCSP_DIRECTIVE imported from ipaserver.install.httpinstance, if import
    succeeds, else ""
  returned: always
  type: str
HTTPD_SSL_CONF:
  description: paths.HTTPD_SSL_CONF from ipaplatform
  returned: always
  type: str
HTTPD_NSS_CONF:
  description: paths.HTTPD_NSS_CONF from ipaplatform
  returned: always
  type: str
HTTPD_ALIAS_DIR:
  description: paths.HTTPD_ALIAS_DIR from ipaplatform
  returned: always
  type: str
allow_httpd_ifp:
  description:
    True if sssd_enable_ifp can be imported from ipaclient.install.client,
    else false.
  returned: always
  type: bool
NSS_DB_DIR:
  description: paths.NSS_DB_DIR from ipaplatform
  returned: always
  type: str
USE_AUTHSELECT:
  description: True if "AUTHSELECT" is defined in paths
  returned: always
  type: bool
python_interpreter:
  description: Python interpreter from sys.executable
  returned: always
  type: str
'''

import sys
from ansible.module_utils.basic import AnsibleModule
try:
    from ipaplatform.paths import paths
    try:
        from ipaserver.install.httpinstance import OCSP_ENABLED, OCSP_DIRECTIVE
        NSS_OCSP_ENABLED = ""
        NSS_OCSP_DIRECTIVE = ""
        NSS_NICKNAME_DIRECTIVE = ""
    except ImportError:
        from ipaserver.install.httpinstance import NSS_OCSP_ENABLED
        NSS_OCSP_DIRECTIVE = "NSSOCSP"
        NSS_NICKNAME_DIRECTIVE = "NSSNickname"
        OCSP_ENABLED = ""
        OCSP_DIRECTIVE = ""
    try:
        from ipaclient.install.client import sssd_enable_ifp
    except ImportError:
        sssd_enable_ifp = None
except ImportError as _err:
    MODULE_IMPORT_ERROR = str(_err)
    paths = None
    sssd_enable_ifp = None
else:
    MODULE_IMPORT_ERROR = None


def main():
    ansible_module = AnsibleModule(
        argument_spec={},
        supports_check_mode=False,
    )

    if MODULE_IMPORT_ERROR is not None:
        ansible_module.fail_json(msg=MODULE_IMPORT_ERROR)

    ansible_module.exit_json(changed=False,
                             NSS_OCSP_ENABLED=NSS_OCSP_ENABLED,
                             NSS_OCSP_DIRECTIVE=NSS_OCSP_DIRECTIVE,
                             NSS_NICKNAME_DIRECTIVE=NSS_NICKNAME_DIRECTIVE,
                             OCSP_ENABLED=OCSP_ENABLED,
                             OCSP_DIRECTIVE=OCSP_DIRECTIVE,
                             HTTPD_SSL_CONF=paths.HTTPD_SSL_CONF,
                             HTTPD_NSS_CONF=paths.HTTPD_NSS_CONF,
                             HTTPD_ALIAS_DIR=paths.HTTPD_ALIAS_DIR,
                             allow_httpd_ifp=sssd_enable_ifp is not None,
                             NSS_DB_DIR=paths.NSS_DB_DIR,
                             USE_AUTHSELECT=hasattr(paths, "AUTHSELECT"),
                             python_interpreter=sys.executable)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""ansible-freeipa iparole module implementation."""

# Authors:
#   Rafael Guterres Jeffman <rjeffman@redhat.com>
#   Thomas Woerner <twoerner@redhat.com>
#
# Copyright (C) 2020-2022 Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {
    "metadata_version": "1.0",
    "supported_by": "community",
    "status": ["preview"],
}


DOCUMENTATION = """
---
module: iparole
short_description: Manage FreeIPA role
description: Manage FreeIPA role
extends_documentation_fragment:
  - ipamodule_base_docs
options:
  name:
    description: The list of role name strings.
    type: list
    elements: str
    required: true
    aliases: ["cn"]
  description:
    description: A description for the role.
    type: str
    required: false
  rename:
    description: Rename the role object.
    type: str
    required: false
    aliases: ["new_name"]
  privilege:
    description: List of privileges
    type: list
    elements: str
    required: false
  user:
    description: List of users.
    type: list
    elements: str
    required: false
  group:
    description: List of groups.
    type: list
    elements: str
    required: false
  host:
    description: List of hosts.
    type: list
    elements: str
    required: false
  hostgroup:
    description: List of hostgroups.
    type: list
    elements: str
    required: false
  service:
    description: List of services.
    type: list
    elements: str
    required: false
  action:
    description: Work on role or member level.
    type: str
    choices: ["role", "member"]
    default: role
    required: false
  state:
    description: The state to ensure.
    type: str
    choices: ["present", "absent", "renamed"]
    default: present
    required: false
author:
  - Rafael Guterres Jeffman (@rjeffman)
  - Thomas Woerner (@t-woerner)
"""

EXAMPLES = """
- name: Ensure a role named `somerole` is present.
  iparole:
    ipaadmin_password: SomeADMINpassword
    name: somerole

- name: Ensure user `pinky` is a memmer of role `somerole`.
  iparole:
    ipaadmin_password: SomeADMINpassword
    name: somerole
    user:
    - pinky
    action: member

- name: Ensure a role named `somerole` is absent.
  iparole:
    ipaadmin_password: SomeADMINpassword
    name: somerole
    state: absent
"""

# pylint: disable=wrong-import-position
# pylint: disable=import-error
# pylint: disable=no-name-in-module
from ansible.module_utils._text import to_text
from ansible.module_utils.ansible_freeipa_module import \
    IPAAnsibleModule, gen_add_del_lists, compare_args_ipa, \
    gen_intersection_list, ensure_fqdn
from ansible.module_utils import six

if six.PY3:
    unicode = str


def find_role(module, name):
    """Find if a role with the given name already exist."""
    try:
        _result = module.ipa_command("role_show", name, {"all": True})
    except Exception:  # pylint: disable=broad-except
        # An exception is raised if role name is not found.
        return None
    else:
        return _result["result"]


def gen_args(module):
    """Generate arguments for executing commands."""
    arg_map = {
        "description": "description",
        "rename": "rename",
    }
    args = {}

    for param, arg in arg_map.items():
        value = module.params_get(param)
        if value is not None:
            args[arg] = value

    return args


def check_parameters(module):
    """Check if parameters passed for module processing are valid."""
    action = module.params_get("action")
    state = module.params_get("state")

    invalid = []

    if state == "renamed":
        if action == "member":
            module.fail_json(
                msg="Invalid action 'member' with state 'renamed'.")
        invalid = [
            "description",
            "user", "group",
            "host", "hostgroup",
            "service",
            "privilege",
        ]

    if state == "present":
        invalid = ["rename"]
        if action == "member":
            invalid.extend(['description'])

    if state == "absent":
        invalid.extend(['description', 'rename'])
        if action != "member":
            invalid.extend(['privilege'])

    module.params_fail_used_invalid(invalid, state, action)


def get_member_host_with_fqdn_lowercase(module, mod_member):
    """Retrieve host members from module, as FQDN, lowercase."""
    default_domain = module.ipa_get_domain()
    hosts = module.params_get(mod_member)
    return (
        [ensure_fqdn(host, default_domain).lower() for host in hosts]
        if hosts
        else hosts
    )


def ensure_absent_state(module, name, action, res_find):
    """Define commands to ensure absent state."""
    commands = []

    if action == "role":
        commands.append([name, 'role_del', {}])

    if action == "member":

        _members = module.params_get_lowercase("privilege")
        if _members is not None:
            del_list = gen_intersection_list(
                _members,
                result_get_value_lowercase(res_find, "memberof_privilege")
            )
            if del_list:
                commands.append([name, "role_remove_privilege",
                                 {"privilege": del_list}])

        member_args = {}
        for key in ['user', 'group', 'hostgroup']:
            _members = module.params_get_lowercase(key)
            if _members:
                del_list = gen_intersection_list(
                    _members,
                    result_get_value_lowercase(res_find, "member_%s" % key)
                )
                if del_list:
                    member_args[key] = del_list

        # ensure hosts are FQDN.
        _members = get_member_host_with_fqdn_lowercase(module, "host")
        if _members:
            del_list = gen_intersection_list(
                _members, res_find.get('member_host'))
            if del_list:
                member_args["host"] = del_list

        _services = get_service_param(module, "service")
        if _services:
            _existing = result_get_value_lowercase(res_find, "member_service")
            items = gen_intersection_list(_services.keys(), _existing)
            if items:
                member_args["service"] = [_services[key] for key in items]

        # Only add remove command if there's at least one member no manage.
        if member_args:
            commands.append([name, "role_remove_member", member_args])

    return commands


def get_service_param(module, key):
    """
    Retrieve dict of services, with realm, from the module parameters.

    As the services are compared in a case insensitive manner, but
    are recorded in a case preserving way, a dict mapping the services
    in lowercase to the provided module parameter is generated, so
    that dict keys can be used for comparison and the values are used
    with IPA API.
    """
    _services = module.params_get(key)
    if _services is not None:
        ipa_realm = module.ipa_get_realm()
        _services = [
            to_text(svc) if '@' in svc else ('%s@%s' % (svc, ipa_realm))
            for svc in _services
        ]
        if _services:
            _services = {svc.lower(): svc for svc in _services}
    return _services


def result_get_value_lowercase(res_find, key, default=None):
    """
    Retrieve a member of a dictionary converted to lowercase.

    If field data is a string it is returned in lowercase. If
    field data is a list or tuple, it is assumed that all values
    are strings and the result is a list of strings in lowercase.

    If 'key' is not found in the dictionary, returns 'default'.
    """
    existing = res_find.get(key)
    if existing is not None:
        if isinstance(existing, (list, tuple)):
            existing = [to_text(item).lower() for item in existing]
        if isinstance(existing, (str, unicode)):
            existing = existing.lower()
    else:
        existing = default
    return existing


def gen_services_add_del_lists(module, mod_member, res_find, res_member):
    """Generate add/del lists for service principals."""
    add_list, del_list = None, None
    _services = get_service_param(module, mod_member)
    if _services is not None:
        _existing = result_get_value_lowercase(res_find, res_member)
        add_list, del_list = gen_add_del_lists(_services.keys(), _existing)
        if add_list:
            add_list = [_services[key] for key in add_list]
        if del_list:
            del_list = [to_text(item) for item in del_list]
    return add_list, del_list


def ensure_role_with_members_is_present(module, name, res_find, action):
    """Define commands to ensure member are present for action `role`."""
    commands = []

    _members = module.params_get_lowercase("privilege")
    if _members:
        add_list, del_list = gen_add_del_lists(
            _members,
            result_get_value_lowercase(res_find, "memberof_privilege")
        )

        if add_list:
            commands.append([name, "role_add_privilege",
                             {"privilege": add_list}])
        if action == "role" and del_list:
            commands.append([name, "role_remove_privilege",
                             {"privilege": del_list}])

    add_members = {}
    del_members = {}

    for key in ["user", "group", "hostgroup"]:
        _members = module.params_get_lowercase(key)
        if _members is not None:
            add_list, del_list = gen_add_del_lists(
                _members,
                result_get_value_lowercase(res_find, "member_%s" % key)
            )
            if add_list:
                add_members[key] = add_list
            if del_list:
                del_members[key] = del_list

    # ensure hosts are FQDN.
    _members = get_member_host_with_fqdn_lowercase(module, "host")
    if _members:
        add_list, del_list = gen_add_del_lists(
            _members, res_find.get('member_host'))
        if add_list:
            add_members["host"] = add_list
        if del_list:
            del_members["host"] = del_list

    (add_services, del_services) = gen_services_add_del_lists(
        module, "service", res_find, "member_service")
    if add_services:
        add_members["service"] = add_services
    if del_services:
        del_members["service"] = del_services

    if add_members:
        commands.append([name, "role_add_member", add_members])
    # Only remove members if ensuring role, not acting on members.
    if action == "role" and del_members:
        commands.append([name, "role_remove_member", del_members])

    return commands


def role_commands_for_name(module, state, action, name):
    """Define commands for the Role module."""
    commands = []

    res_find = find_role(module, name)

    if state == "renamed":
        args = gen_args(module)
        if res_find is None:
            module.fail_json(msg="No role '%s'" % name)
        else:
            commands.append([name, 'role_mod', args])

    if state == "present":
        args = gen_args(module)

        if action == "role":
            if res_find is None:
                commands.append([name, 'role_add', args])
                res_find = {}
            else:
                if not compare_args_ipa(module, args, res_find):
                    commands.append([name, 'role_mod', args])

        if action == "member":
            if res_find is None:
                module.fail_json(msg="No role '%s'" % name)

        cmds = ensure_role_with_members_is_present(
            module, name, res_find, action
        )
        commands.extend(cmds)

    if state == "absent" and res_find is not None:
        cmds = ensure_absent_state(module, name, action, res_find)
        commands.extend(cmds)

    return commands


def create_module():
    """Create module description."""
    ansible_module = IPAAnsibleModule(
        argument_spec=dict(
            # generalgroups
            name=dict(type="list", elements="str", aliases=["cn"],
                      required=True),
            # present
            description=dict(required=False, type="str", default=None),
            rename=dict(required=False, type="str", default=None,
                        aliases=["new_name"]),
            # members
            privilege=dict(required=False, type='list', elements="str",
                           default=None),
            user=dict(required=False, type='list', elements="str",
                      default=None),
            group=dict(required=False, type='list', elements="str",
                       default=None),
            host=dict(required=False, type='list', elements="str",
                      default=None),
            hostgroup=dict(required=False, type='list', elements="str",
                           default=None),
            service=dict(required=False, type='list', elements="str",
                         default=None),

            # state
            action=dict(type="str", default="role",
                        choices=["role", "member"]),
            state=dict(type="str", default="present",
                       choices=["present", "absent", "renamed"]),
        ),
        supports_check_mode=True,
        mutually_exclusive=[],
        required_one_of=[]
    )

    ansible_module._ansible_debug = True  # pylint: disable=protected-access

    return ansible_module


def main():
    """Process role module script."""
    ansible_module = create_module()
    check_parameters(ansible_module)

    # Init

    # Connect to IPA API
    with ansible_module.ipa_connect():

        state = ansible_module.params_get("state")
        action = ansible_module.params_get("action")
        names = ansible_module.params_get("name")
        commands = []

        for name in names:
            cmds = role_commands_for_name(ansible_module, state, action, name)
            commands.extend(cmds)

        exit_args = {}

        # Execute commands

        changed = ansible_module.execute_ipa_commands(
            commands, fail_on_member_errors=True)

    # Done
    ansible_module.exit_json(changed=changed, **exit_args)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Based on ipa-client-install code
#
# Copyright (C) 2017-2022  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

__all__ = ["gssapi", "version", "ipadiscovery", "api", "errors", "x509",
           "constants", "sysrestore", "certmonger", "certstore",
           "delete_persistent_client_session_data", "ScriptError",
           "CheckedIPAddress", "validate_domain_name", "normalize_hostname",
           "validate_hostname", "services", "tasks", "CalledProcessError",
           "write_tmp_file", "ipa_generate_password", "DN", "kinit_keytab",
           "kinit_password", "GSSError", "CLIENT_INSTALL_ERROR",
           "is_ipa_client_installed", "CLIENT_ALREADY_CONFIGURED",
           "nssldap_exists", "remove_file", "check_ip_addresses",
           "print_port_conf_info", "configure_ipa_conf", "purge_host_keytab",
           "configure_sssd_conf", "realm_to_suffix", "run", "timeconf",
           "serialization", "configure_krb5_conf", "get_ca_certs",
           "SECURE_PATH", "get_server_connection_interface",
           "disable_ra", "client_dns",
           "configure_certmonger", "update_ssh_keys",
           "configure_openldap_conf", "hardcode_ldap_server",
           "get_certs_from_ldap", "save_state", "create_ipa_nssdb",
           "configure_nisdomain", "configure_ldap_conf",
           "configure_nslcd_conf", "configure_ssh_config",
           "configure_sshd_config", "configure_automount",
           "configure_firefox", "sync_time", "check_ldap_conf",
           "sssd_enable_ifp", "getargspec", "paths", "options",
           "IPA_PYTHON_VERSION", "NUM_VERSION", "certdb", "get_ca_cert",
           "ipalib", "logger", "ipautil", "installer"]

import sys

# Import getargspec from inspect or provide own getargspec for
# Python 2 compatibility with Python 3.11+.
try:
    from inspect import getargspec
except ImportError:
    from collections import namedtuple
    from inspect import getfullargspec

    # The code is copied from Python 3.10 inspect.py
    # Authors: Ka-Ping Yee <ping@lfw.org>
    #          Yury Selivanov <yselivanov@sprymix.com>
    ArgSpec = namedtuple('ArgSpec', 'args varargs keywords defaults')

    def getargspec(func):
        args, varargs, varkw, defaults, kwonlyargs, _kwonlydefaults, \
            ann = getfullargspec(func)
        if kwonlyargs or ann:
            raise ValueError(
                "Function has keyword-only parameters or annotations"
                ", use inspect.signature() API which can support them")
        return ArgSpec(args, varargs, varkw, defaults)


try:
    from ipapython.version import NUM_VERSION, VERSION

    if NUM_VERSION < 30201:
        # See ipapython/version.py
        IPA_MAJOR, IPA_MINOR, IPA_RELEASE = [int(x) for x in
                                             VERSION.split(".", 2)]
        IPA_PYTHON_VERSION = IPA_MAJOR * 10000 + IPA_MINOR * 100 + IPA_RELEASE
    else:
        IPA_PYTHON_VERSION = NUM_VERSION

    # pylint: disable=invalid-name,useless-object-inheritance
    class installer_obj(object):
        def __init__(self):
            pass

        # pylint: disable=attribute-defined-outside-init
        def set_logger(self, _logger):
            self.logger = _logger

        # def __getattribute__(self, attr):
        #    value = super(installer_obj, self).__getattribute__(attr)
        #    if not attr.startswith("--") and not attr.endswith("--"):
        #        logger.debug(
        #            "  <-- Accessing installer.%s (%s)" % (attr, repr(value)))
        #    return value

        # def __getattr__(self, attr):
        #    # logger.info("  --> ADDING missing installer.%s" % attr)
        #    self.logger.warn("  --> ADDING missing installer.%s" % attr)
        #    setattr(self, attr, None)
        #    return getattr(self, attr)

        # def __setattr__(self, attr, value):
        #    logger.debug("  --> Setting installer.%s to %s" %
        #                 (attr, repr(value)))
        #    return super(installer_obj, self).__setattr__(attr, value)

        def knobs(self):
            for name in self.__dict__:
                yield self, name

    # Initialize installer settings
    installer = installer_obj()
    # Create options
    options = installer
    # pylint: disable=attribute-defined-outside-init
    options.interactive = False
    options.unattended = not options.interactive

    if NUM_VERSION >= 40400:
        # IPA version >= 4.4

        # import sys
        import gssapi
        import logging

        from ipapython import version
        try:
            from ipaclient.install import ipadiscovery
        except ImportError:
            from ipaclient import ipadiscovery
        import ipalib
        from ipalib import api, errors, x509
        from ipalib import constants
        try:
            from ipalib import sysrestore
        except ImportError:
            try:
                from ipalib.install import sysrestore
            except ImportError:
                from ipapython import sysrestore
        try:
            from ipalib.install import certmonger
        except ImportError:
            from ipapython import certmonger
        try:
            from ipalib.install import certstore
        except ImportError:
            from ipalib import certstore
        from ipalib.rpc import delete_persistent_client_session_data
        from ipapython import certdb, ipautil
        from ipapython.admintool import ScriptError
        from ipapython.ipautil import CheckedIPAddress
        from ipalib.util import validate_domain_name, normalize_hostname, \
            validate_hostname
        from ipaplatform import services
        from ipaplatform.paths import paths
        from ipaplatform.tasks import tasks
        try:
            from cryptography.hazmat.primitives import serialization
        except ImportError:
            serialization = None
        from ipapython.ipautil import CalledProcessError, write_tmp_file, \
            ipa_generate_password
        from ipapython.dn import DN
        try:
            from ipalib.install.kinit import kinit_keytab, kinit_password
        except ImportError:
            from ipapython.ipautil import kinit_keytab, kinit_password
        from ipapython.ipa_log_manager import standard_logging_setup
        from gssapi.exceptions import GSSError
        try:
            from ipaclient.install.client import configure_krb5_conf, \
                get_ca_certs, SECURE_PATH, get_server_connection_interface, \
                disable_ra, client_dns, \
                configure_certmonger, update_ssh_keys, \
                configure_openldap_conf, \
                hardcode_ldap_server, get_certs_from_ldap, save_state, \
                create_ipa_nssdb, configure_ssh_config, \
                configure_sshd_config, \
                configure_automount, configure_firefox, configure_nisdomain, \
                CLIENT_INSTALL_ERROR, is_ipa_client_installed, \
                CLIENT_ALREADY_CONFIGURED, nssldap_exists, remove_file, \
                check_ip_addresses, print_port_conf_info, configure_ipa_conf, \
                purge_host_keytab, configure_sssd_conf, configure_ldap_conf, \
                configure_nslcd_conf
            get_ca_cert = None
        except ImportError:
            # Create temporary copy of ipa-client-install script (as
            # ipa_client_install.py) to be able to import the script easily
            # and also to remove the global finally clause in which the
            # generated ccache file gets removed. The ccache file will be
            # needed in the next step.
            # This is done in a temporary directory that gets removed right
            # after ipa_client_install has been imported.
            import shutil
            import tempfile
            temp_dir = tempfile.mkdtemp(dir="/tmp")
            sys.path.append(temp_dir)
            temp_file = "%s/ipa_client_install.py" % temp_dir

            with open("/usr/sbin/ipa-client-install", "r") as f_in:
                with open(temp_file, "w") as f_out:
                    for line in f_in:
                        if line.startswith("finally:"):
                            break
                        f_out.write(line)
            import ipa_client_install

            shutil.rmtree(temp_dir, ignore_errors=True)
            sys.path.remove(temp_dir)

            # pylint: disable=deprecated-method
            argspec = getargspec(
                ipa_client_install.configure_krb5_conf)
            if argspec.keywords is None:
                def configure_krb5_conf(
                        cli_realm, cli_domain, cli_server, cli_kdc, dnsok,
                        filename, client_domain, client_hostname, force=False,
                        configure_sssd=True):
                    # pylint: disable=global-variable-not-assigned
                    global options
                    options.force = force
                    options.sssd = configure_sssd
                    return ipa_client_install.configure_krb5_conf(
                        cli_realm, cli_domain, cli_server, cli_kdc, dnsok,
                        options, filename, client_domain, client_hostname)
            else:
                configure_krb5_conf = ipa_client_install.configure_krb5_conf
            if NUM_VERSION < 40100:
                get_ca_cert = ipa_client_install.get_ca_cert
                get_ca_certs = None
            else:
                get_ca_cert = None
                get_ca_certs = ipa_client_install.get_ca_certs
            SECURE_PATH = ("/bin:/sbin:/usr/kerberos/bin:/usr/kerberos/sbin:"
                           "/usr/bin:/usr/sbin")

            get_server_connection_interface = \
                ipa_client_install.get_server_connection_interface
            disable_ra = ipa_client_install.disable_ra
            client_dns = ipa_client_install.client_dns
            configure_certmonger = ipa_client_install.configure_certmonger
            update_ssh_keys = ipa_client_install.update_ssh_keys
            configure_openldap_conf = \
                ipa_client_install.configure_openldap_conf
            hardcode_ldap_server = ipa_client_install.hardcode_ldap_server
            get_certs_from_ldap = ipa_client_install.get_certs_from_ldap
            save_state = ipa_client_install.save_state

            create_ipa_nssdb = certdb.create_ipa_nssdb

            argspec = \
                getargspec(ipa_client_install.configure_nisdomain)
            if len(argspec.args) == 3:
                configure_nisdomain = ipa_client_install.configure_nisdomain
            else:
                def configure_nisdomain(_options, domain, _statestore=None):
                    return ipa_client_install.configure_nisdomain(_options,
                                                                  domain)

            configure_ldap_conf = ipa_client_install.configure_ldap_conf
            configure_nslcd_conf = ipa_client_install.configure_nslcd_conf

            configure_ssh_config = ipa_client_install.configure_ssh_config
            configure_sshd_config = ipa_client_install.configure_sshd_config
            configure_automount = ipa_client_install.configure_automount
            configure_firefox = ipa_client_install.configure_firefox

        from ipapython.ipautil import realm_to_suffix, run

        try:
            from ipaclient.install import timeconf
            time_service = "chronyd"
        except ImportError:
            try:
                from ipaclient.install import ntpconf as timeconf
            except ImportError:
                from ipaclient import ntpconf as timeconf
            time_service = "ntpd"

        try:
            from ipaclient.install.client import sync_time
        except ImportError:
            sync_time = None

        try:
            from ipaclient.install.client import check_ldap_conf
        except ImportError:
            check_ldap_conf = None

        try:
            from ipaclient.install.client import sssd_enable_ifp
        except ImportError:
            sssd_enable_ifp = None

        logger = logging.getLogger("ipa-client-install")
        root_logger = logger

    else:
        # IPA version < 4.4

        raise Exception("freeipa version '%s' is too old" % VERSION)

except ImportError as _err:
    ANSIBLE_IPA_CLIENT_MODULE_IMPORT_ERROR = str(_err)

    for attr in __all__:
        setattr(sys.modules[__name__], attr, None)

else:
    ANSIBLE_IPA_CLIENT_MODULE_IMPORT_ERROR = None


def setup_logging():
    standard_logging_setup(
        paths.IPACLIENT_INSTALL_LOG, verbose=False, debug=False,
        filemode='a', console_format='%(message)s')


def ansible_module_get_parsed_ip_addresses(ansible_module,
                                           param='ip_addresses'):
    ip_addresses = ansible_module.params.get(param)
    if ip_addresses is None:
        return None

    ip_addrs = []
    for ip in ip_addresses:
        try:
            ip_parsed = ipautil.CheckedIPAddress(ip)
        except Exception as e:
            ansible_module.fail_json(msg="Invalid IP Address %s: %s" % (ip, e))
        ip_addrs.append(ip_parsed)
    return ip_addrs


def check_imports(module):
    if ANSIBLE_IPA_CLIENT_MODULE_IMPORT_ERROR is not None:
        module.fail_json(msg=ANSIBLE_IPA_CLIENT_MODULE_IMPORT_ERROR)



#!/usr/bin/python

"""Check which tests are scheduled to be executed."""

import sys
import re
import os
import yaml


RE_IS_TEST = re.compile(r"(.*/)?test_.*\.yml")
RE_IS_VARS = re.compile(r"(.*/)?variables(_.*)?\.yaml")
REPO_ROOT = os.path.join(os.path.dirname(__file__), "..")


def get_tests():
    """Retrieve a list of modules and its tests."""

    def get_module(root):
        if root != _test_dir:
            while True:
                module = os.path.basename(root)
                root = os.path.dirname(root)
                if root == _test_dir:
                    return module
        return "."

    _result = {}
    _test_dir = os.path.join(REPO_ROOT, "tests")
    for root, _dirs, files in os.walk(_test_dir):
        module = get_module(root)
        _result[module] = [
            os.path.splitext(test)[0]
            for test in files
            if RE_IS_TEST.search(test)
        ]

    return _result


def get_test_config(scenarios):
    template_path = os.path.join(REPO_ROOT, "tests/azure/templates")
    _result = {}
    for _root, _dirs, files in os.walk(template_path):
        for filename in [x for x in files if RE_IS_VARS.search(x)]:
            _templates, *scenario = os.path.basename(
                os.path.splitext(filename)[0]
            ).split("_", 1)
            scenario = scenario[0] if scenario else "All"
            _result[scenario] = {}
            # only process selected scenarios
            if scenario not in scenarios and len(scenarios) > 1:
                continue
            with open(os.path.join(template_path, filename), "rt") as inp:
                data = yaml.safe_load(inp)
            if not data["variables"].get("empty", False):
                variables = data["variables"]
                for key, value in variables.items():
                    variables[key] = [
                        x.strip() for x in value.split(",") if x.strip()
                    ]
                _result[scenario] = variables

    return _result


def print_configuration(scenario, disabled, enabled):
    """Print the test configuration for a scenario."""
    print(f"\nScenario: {scenario}")
    for test_cfg, title in [(disabled, "Disabled"), (enabled, "Enabled")]:
        print(f"    {title} tests:")
        if test_cfg:
            for module, tests in test_cfg.items():
                print(f"        {module}:")
                for test in tests:
                    print(f"            - {test}")
        else:
            print("        No custom configuration.")


def main():
    if any(item in sys.argv for item in ["-h", "--help"]):
        print("usage: check_test_config.py [-h|--help] [SCENARIO...]")
        return

    scenarios = ["All"] + sys.argv[1:]
    all_tests = get_tests()
    test_config = get_test_config(scenarios)

    print("Test configuration:")
    for scenario in sorted(test_config.keys()):
        if scenario not in scenarios and len(scenarios) > 1:
            continue
        # extract scenario configuration
        config = test_config[scenario]
        disabled = {}
        enabled = {}
        for res, state in [(disabled, "disabled"), (enabled, "enabled")]:
            items = [
                x.strip()
                for x in
                os.environ.get(f"ipa_{state}_modules".upper(), "").split(",")
                if x.strip()
            ] if scenario == "All" else []
            modules = config.get(f"ipa_{state}_modules", []) + items
            for module in modules:
                if module != "None":
                    res[module] = set(all_tests[module])
            items = [
                x.strip()
                for x in
                os.environ.get(f"ipa_{state}_tests".upper(), "").split(",")
                if x.strip()
            ] if scenario == "All" else []
            test_list = config.get(f"ipa_{state}_tests", []) + items
            for test in test_list:
                if test == "None":
                    continue
                for module, tests in all_tests.items():
                    if test in tests:
                        mod = res.setdefault(module, set())
                        mod.add(test)
                        tests.remove(test)
        print_configuration(scenario, disabled, enabled)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Authors:
#   Chris Procter <cprocter@redhat.com>
#   Thomas Woerner <twoerner@redhat.com>
#
# Copyright (C) 2021-2022 Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {
    "metadata_version": "1.0",
    "supported_by": "community",
    "status": ["preview"],
}


DOCUMENTATION = '''
---
module: ipaautomountkey
author:
  - Chris Procter (@chr15p))
  - Thomas Woerner (@t-woerner)
short_description: Manage FreeIPA autommount map
description:
- Add, delete, and modify an IPA automount map
extends_documentation_fragment:
  - ipamodule_base_docs
options:
  location:
    description: automount location map is in
    type: str
    required: True
    aliases: ["automountlocationcn", "automountlocation"]
  mapname:
    description: automount map to be managed
    type: str
    aliases: ["map", "automountmapname", "automountmap"]
    required: True
  key:
    description: automount key to be managed
    type: str
    required: True
    aliases: ["name", "automountkey"]
  rename:
    description: key to change to if state is 'renamed'
    type: str
    required: False
    aliases: ["new_name", "newautomountkey"]
  info:
    description: Mount information for the key
    type: str
    required: False
    aliases: ["information", "automountinformation"]
  state:
    description: State to ensure
    type: str
    required: False
    default: present
    choices: ["present", "absent", "renamed"]
'''

EXAMPLES = '''
  - name: create key TestKey
    ipaautomountkey:
      ipaadmin_password: SomeADMINpassword
      locationcn: TestLocation
      mapname: TestMap
      key: TestKey
      info: 192.168.122.1:/exports
      state: present

  - name: ensure key TestKey is absent
    ipaautomountkey:
      ipaadmin_password: SomeADMINpassword
      location: TestLocation
      mapname: TestMap
      key: TestKey
      state: absent
'''

RETURN = '''
'''

from ansible.module_utils.ansible_freeipa_module import (
    IPAAnsibleModule, ipalib_errors
)


class AutomountKey(IPAAnsibleModule):

    def __init__(self, *args, **kwargs):
        # pylint: disable=super-with-arguments
        super(AutomountKey, self).__init__(*args, **kwargs)
        self.commands = []

    def get_key(self, location, mapname, key):
        try:
            args = {
                "automountmapautomountmapname": mapname,
                "automountkey": key,
                "all": True,
            }
            resp = self.ipa_command("automountkey_show", location, args)
        except ipalib_errors.NotFound:
            return None
        else:
            return resp.get("result")

    def check_ipa_params(self):
        invalid = []
        state = self.params_get("state")
        if state == "present":
            invalid = ["rename"]
            if not self.params_get("info"):
                self.fail_json(msg="Value required for argument 'info'")

        if state == "rename":
            invalid = ["info"]
            if not self.params_get("rename"):
                self.fail_json(msg="Value required for argument 'renamed'")

        if state == "absent":
            invalid = ["info", "rename"]

        self.params_fail_used_invalid(invalid, state)

    @staticmethod
    def get_args(mapname, key, info, rename):
        _args = {}
        if mapname:
            _args["automountmapautomountmapname"] = mapname
        if key:
            _args["automountkey"] = key
        if info:
            _args["automountinformation"] = info
        if rename:
            _args["rename"] = rename
        return _args

    def define_ipa_commands(self):
        state = self.params_get("state")
        location = self.params_get("location")
        mapname = self.params_get("mapname")
        key = self.params_get("key")
        info = self.params_get("info")
        rename = self.params_get("rename")

        args = self.get_args(mapname, key, info, rename)

        res_find = self.get_key(location, mapname, key)

        if state == "present":
            if res_find is None:
                # does not exist and is wanted
                self.commands.append([location, "automountkey_add", args])
            else:
                # exists and is wanted, check for changes
                if info not in res_find.get("automountinformation"):
                    self.commands.append([location, "automountkey_mod", args])

        if state == "renamed":
            if res_find is None:
                self.fail_json(
                    msg=(
                        "Cannot rename inexistent key: '%s', '%s', '%s'"
                        % (location, mapname, key)
                    )
                )
            self.commands.append([location, "automountkey_mod", args])

        if state == "absent":
            # if key exists and self.ipa_params.state == "absent":
            if res_find is not None:
                self.commands.append([location, "automountkey_del", args])


def main():
    ipa_module = AutomountKey(
        argument_spec=dict(
            state=dict(
                type='str',
                choices=['present', 'absent', 'renamed'],
                required=False,
                default='present',
            ),
            location=dict(
                type="str",
                aliases=["automountlocationcn", "automountlocation"],
                required=True,
            ),
            rename=dict(
                type="str",
                aliases=["new_name", "newautomountkey"],
                required=False,
            ),
            mapname=dict(
                type="str",
                aliases=["map", "automountmapname", "automountmap"],
                required=True,
            ),
            key=dict(
                type="str",
                aliases=["name", "automountkey"],
                required=True,
                no_log=False,
            ),
            info=dict(
                type="str",
                aliases=["information", "automountinformation"],
                required=False,
            ),
        ),
    )
    ipaapi_context = ipa_module.params_get("ipaapi_context")
    with ipa_module.ipa_connect(context=ipaapi_context):
        ipa_module.check_ipa_params()
        ipa_module.define_ipa_commands()
        changed = ipa_module.execute_ipa_commands(ipa_module.commands)
    ipa_module.exit_json(changed=changed)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Copyright (C) 2021-2022  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.0',
    'supported_by': 'community',
    'status': ['preview'],
}

DOCUMENTATION = '''
---
module: ipabackup_get_backup_dir
short_description:
  Get IPA_BACKUP_DIR from ipaplatform
description:
  Get IPA_BACKUP_DIR from ipaplatform
author:
    - Thomas Woerner (@t-woerner)
'''

EXAMPLES = '''
# Get IPA_BACKUP_DIR from ipaplatform
- name: Get IPA_BACKUP_DIR from ipaplatform
  ipabackup_get_backup_dir:
  register: result
'''

RETURN = '''
backup_dir:
  description: IPA_BACKUP_DIR from ipaplatform
  returned: always
  type: str
'''

from ansible.module_utils.basic import AnsibleModule
try:
    from ipaplatform.paths import paths
except ImportError as _err:
    MODULE_IMPORT_ERROR = str(_err)
    paths = None
else:
    MODULE_IMPORT_ERROR = None


def main():
    module = AnsibleModule(
        argument_spec={},
        supports_check_mode=True,
    )

    if MODULE_IMPORT_ERROR is not None:
        module.fail_json(msg=MODULE_IMPORT_ERROR)

    module.exit_json(changed=False,
                     backup_dir=paths.IPA_BACKUP_DIR)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Based on ipa-client-install code
#
# Copyright (C) 2018-2022  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.0',
    'supported_by': 'community',
    'status': ['preview'],
}

DOCUMENTATION = '''
---
module: ipaclient_set_hostname
short_description: Backup and set hostname
description:
  Backup and set hostname
options:
  hostname:
    description: Fully qualified name of this host
    type: str
    required: yes
author:
    - Thomas Woerner (@t-woerner)
'''

EXAMPLES = '''
# Backup and set hostname
- name: Backup and set hostname
  ipaclient_set_hostname:
    hostname: client1.example.com
'''

RETURN = '''
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.ansible_ipa_client import (
    setup_logging, check_imports, sysrestore, paths, tasks
)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            hostname=dict(required=True, type='str'),
        ),
        supports_check_mode=False,
    )

    module._ansible_debug = True
    check_imports(module)
    setup_logging()

    hostname = module.params.get('hostname')

    fstore = sysrestore.FileStore(paths.IPA_CLIENT_SYSRESTORE)
    statestore = sysrestore.StateFile(paths.IPA_CLIENT_SYSRESTORE)

    tasks.backup_hostname(fstore, statestore)
    tasks.set_hostname(hostname)

    module.exit_json(changed=True)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Copyright (C) 2019-2022 Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {
    "metadata_version": "1.0",
    "supported_by": "community",
    "status": ["preview"],
}

DOCUMENTATION = """
---
module: ipahost
short_description: Manage FreeIPA hosts
description: Manage FreeIPA hosts
extends_documentation_fragment:
  - ipamodule_base_docs
options:
  name:
    description: The full qualified domain name.
    type: list
    elements: str
    aliases: ["fqdn"]
    required: false
  hosts:
    description: The list of user host dicts
    required: false
    type: list
    elements: dict
    suboptions:
      name:
        description: The host (internally uid).
        type: str
        aliases: ["fqdn"]
        required: true
      description:
        description: The host description
        type: str
        required: false
      locality:
        description: Host locality (e.g. "Baltimore, MD")
        type: str
        required: false
      location:
        description: Host location (e.g. "Lab 2")
        type: str
        aliases: ["ns_host_location"]
        required: false
      platform:
        description: Host hardware platform (e.g. "Lenovo T61")
        type: str
        aliases: ["ns_hardware_platform"]
        required: false
      os:
        description: Host operating system and version (e.g. "Fedora 9")
        type: str
        aliases: ["ns_os_version"]
        required: false
      password:
        description: Password used in bulk enrollment
        type: str
        aliases: ["user_password", "userpassword"]
        required: false
      random:
        description:
          Initiate the generation of a random password to be used in bulk
          enrollment
        type: bool
        aliases: ["random_password"]
        required: false
      certificate:
        description: List of base-64 encoded host certificates
        type: list
        elements: str
        aliases: ["usercertificate"]
        required: false
      managedby_host:
        description: List of hosts that can manage this host
        type: list
        elements: str
        required: false
      principal:
        description: List of principal aliases for this host
        type: list
        elements: str
        aliases: ["principalname", "krbprincipalname"]
        required: false
      allow_create_keytab_user:
        description: Users allowed to create a keytab of this host
        type: list
        elements: str
        aliases: ["ipaallowedtoperform_write_keys_user"]
        required: false
      allow_create_keytab_group:
        description: Groups allowed to create a keytab of this host
        type: list
        elements: str
        aliases: ["ipaallowedtoperform_write_keys_group"]
        required: false
      allow_create_keytab_host:
        description: Hosts allowed to create a keytab of this host
        type: list
        elements: str
        aliases: ["ipaallowedtoperform_write_keys_host"]
        required: false
      allow_create_keytab_hostgroup:
        description: Hostgroups allowed to create a keytab of this host
        type: list
        elements: str
        aliases: ["ipaallowedtoperform_write_keys_hostgroup"]
        required: false
      allow_retrieve_keytab_user:
        description: Users allowed to retrieve a keytab of this host
        type: list
        elements: str
        aliases: ["ipaallowedtoperform_read_keys_user"]
        required: false
      allow_retrieve_keytab_group:
        description: Groups allowed to retrieve a keytab of this host
        type: list
        elements: str
        aliases: ["ipaallowedtoperform_read_keys_group"]
        required: false
      allow_retrieve_keytab_host:
        description: Hosts allowed to retrieve a keytab of this host
        type: list
        elements: str
        aliases: ["ipaallowedtoperform_read_keys_host"]
        required: false
      allow_retrieve_keytab_hostgroup:
        description: Hostgroups allowed to retrieve a keytab of this host
        type: list
        elements: str
        aliases: ["ipaallowedtoperform_read_keys_hostgroup"]
        required: false
      mac_address:
        description: List of hardware MAC addresses.
        type: list
        elements: str
        aliases: ["macaddress"]
        required: false
      sshpubkey:
        description: List of SSH public keys
        type: list
        elements: str
        aliases: ["ipasshpubkey"]
        required: false
      userclass:
        description:
          Host category (semantics placed on this attribute are for local
          interpretation)
        type: list
        elements: str
        aliases: ["class"]
        required: false
      auth_ind:
        description:
          Defines an allow list for Authentication Indicators. Use 'otp'
          to allow OTP-based 2FA authentications. Use 'radius' to allow
          RADIUS-based 2FA authentications. Other values may be used
          for custom configurations. Use empty string to reset auth_ind
          to the initial value.
        type: list
        elements: str
        aliases: ["krbprincipalauthind"]
        choices: ["radius", "otp", "pkinit", "hardened", ""]
        required: false
      requires_pre_auth:
        description: Pre-authentication is required for the service
        type: bool
        aliases: ["ipakrbrequirespreauth"]
        required: false
      ok_as_delegate:
        description: Client credentials may be delegated to the service
        type: bool
        aliases: ["ipakrbokasdelegate"]
        required: false
      ok_to_auth_as_delegate:
        description:
          The service is allowed to authenticate on behalf of a client
        type: bool
        aliases: ["ipakrboktoauthasdelegate"]
        required: false
      force:
        description: Force host name even if not in DNS
        type: bool
        required: false
      reverse:
        description: Reverse DNS detection
        type: bool
        required: false
      ip_address:
        description:
          The host IP address list (IPv4 and IPv6). No IP address conflict
          check will be done.
        type: list
        elements: str
        aliases: ["ipaddress"]
        required: false
      update_dns:
        description:
          Controls the update of the DNS SSHFP records for existing hosts and
          the removal of all DNS entries if a host gets removed with state
          absent.
        type: bool
        aliases: ["updatedns"]
        required: false
  description:
    description: The host description
    type: str
    required: false
  locality:
    description: Host locality (e.g. "Baltimore, MD")
    type: str
    required: false
  location:
    description: Host location (e.g. "Lab 2")
    type: str
    aliases: ["ns_host_location"]
    required: false
  platform:
    description: Host hardware platform (e.g. "Lenovo T61")
    type: str
    aliases: ["ns_hardware_platform"]
    required: false
  os:
    description: Host operating system and version (e.g. "Fedora 9")
    type: str
    aliases: ["ns_os_version"]
    required: false
  password:
    description: Password used in bulk enrollment
    type: str
    aliases: ["user_password", "userpassword"]
    required: false
  random:
    description:
      Initiate the generation of a random password to be used in bulk
      enrollment
    type: bool
    aliases: ["random_password"]
    required: false
  certificate:
    description: List of base-64 encoded host certificates
    type: list
    elements: str
    aliases: ["usercertificate"]
    required: false
  managedby_host:
    description: List of hosts that can manage this host
    type: list
    elements: str
    required: false
  principal:
    description: List of principal aliases for this host
    type: list
    elements: str
    aliases: ["principalname", "krbprincipalname"]
    required: false
  allow_create_keytab_user:
    description: Users allowed to create a keytab of this host
    type: list
    elements: str
    aliases: ["ipaallowedtoperform_write_keys_user"]
    required: false
  allow_create_keytab_group:
    description: Groups allowed to create a keytab of this host
    type: list
    elements: str
    aliases: ["ipaallowedtoperform_write_keys_group"]
    required: false
  allow_create_keytab_host:
    description: Hosts allowed to create a keytab of this host
    type: list
    elements: str
    aliases: ["ipaallowedtoperform_write_keys_host"]
    required: false
  allow_create_keytab_hostgroup:
    description: Hostgroups allowed to create a keytab of this host
    type: list
    elements: str
    aliases: ["ipaallowedtoperform_write_keys_hostgroup"]
    required: false
  allow_retrieve_keytab_user:
    description: Users allowed to retrieve a keytab of this host
    type: list
    elements: str
    aliases: ["ipaallowedtoperform_read_keys_user"]
    required: false
  allow_retrieve_keytab_group:
    description: Groups allowed to retrieve a keytab of this host
    type: list
    elements: str
    aliases: ["ipaallowedtoperform_read_keys_group"]
    required: false
  allow_retrieve_keytab_host:
    description: Hosts allowed to retrieve a keytab of this host
    type: list
    elements: str
    aliases: ["ipaallowedtoperform_read_keys_host"]
    required: false
  allow_retrieve_keytab_hostgroup:
    description: Hostgroups allowed to retrieve a keytab of this host
    type: list
    elements: str
    aliases: ["ipaallowedtoperform_read_keys_hostgroup"]
    required: false
  mac_address:
    description: List of hardware MAC addresses.
    type: list
    elements: str
    aliases: ["macaddress"]
    required: false
  sshpubkey:
    description: List of SSH public keys
    type: list
    elements: str
    aliases: ["ipasshpubkey"]
    required: false
  userclass:
    description:
      Host category (semantics placed on this attribute are for local
      interpretation)
    type: list
    elements: str
    aliases: ["class"]
    required: false
  auth_ind:
    description:
      Defines an allow list for Authentication Indicators. Use 'otp'
      to allow OTP-based 2FA authentications. Use 'radius' to allow
      RADIUS-based 2FA authentications. Other values may be used
      for custom configurations. Use empty string to reset auth_ind
      to the initial value.
    type: list
    elements: str
    aliases: ["krbprincipalauthind"]
    choices: ["radius", "otp", "pkinit", "hardened", ""]
    required: false
  requires_pre_auth:
    description: Pre-authentication is required for the service
    type: bool
    aliases: ["ipakrbrequirespreauth"]
    required: false
  ok_as_delegate:
    description: Client credentials may be delegated to the service
    type: bool
    aliases: ["ipakrbokasdelegate"]
    required: false
  ok_to_auth_as_delegate:
    description:
      The service is allowed to authenticate on behalf of a client
    type: bool
    aliases: ["ipakrboktoauthasdelegate"]
    required: false
  force:
    description: Force host name even if not in DNS
    type: bool
    required: false
  reverse:
    description: Reverse DNS detection
    type: bool
    required: false
  ip_address:
    description:
      The host IP address list (IPv4 and IPv6). No IP address conflict
      check will be done.
    type: list
    elements: str
    aliases: ["ipaddress"]
    required: false
  update_dns:
    description:
      Controls the update of the DNS SSHFP records for existing hosts and
      the removal of all DNS entries if a host gets removed with state
      absent.
    type: bool
    aliases: ["updatedns"]
    required: false
  update_password:
    description:
      Set password for a host in present state only on creation or always
    type: str
    choices: ["always", "on_create"]
  action:
    description: Work on host or member level
    type: str
    default: "host"
    choices: ["member", "host"]
  state:
    description: State to ensure
    type: str
    default: present
    choices: ["present", "absent",
              "disabled"]
author:
  - Thomas Woerner (@t-woerner)
"""

EXAMPLES = """
# Ensure host is present
- ipahost:
    ipaadmin_password: SomeADMINpassword
    name: host01.example.com
    description: Example host
    ip_address: 192.168.0.123
    locality: Lab
    ns_host_location: Lab
    ns_os_version: CentOS 7
    ns_hardware_platform: Lenovo T61
    mac_address:
    - "08:00:27:E3:B1:2D"
    - "52:54:00:BD:97:1E"
    state: present

# Ensure host is present without DNS
- ipahost:
    ipaadmin_password: SomeADMINpassword
    name: host02.example.com
    description: Example host
    force: yes

# Initiate generation of a random password for the host
- ipahost:
    ipaadmin_password: SomeADMINpassword
    name: host01.example.com
    description: Example host
    ip_address: 192.168.0.123
    random: yes

# Ensure host is disabled
- ipahost:
    ipaadmin_password: SomeADMINpassword
    name: host01.example.com
    update_dns: yes
    state: disabled

# Ensure host is absent
- ipahost:
    ipaadmin_password: SomeADMINpassword
    name: host01.example.com
    state: absent
"""

RETURN = """
host:
  description: Host dict with random password
  returned: If random is yes and user did not exist or update_password is yes
  type: dict
  contains:
    randompassword:
      description: The generated random password
      type: str
      returned: If only one user is handled by the module
    name:
      description: The user name of the user that got a new random password
      returned: If several users are handled by the module
      type: dict
      contains:
        randompassword:
          description: The generated random password
          type: str
          returned: always
"""

from ansible.module_utils.ansible_freeipa_module import \
    IPAAnsibleModule, compare_args_ipa, gen_add_del_lists, \
    encode_certificate, is_ipv4_addr, is_ipv6_addr, ipalib_errors
from ansible.module_utils import six
if six.PY3:
    unicode = str


def find_host(module, name):
    _args = {
        "all": True,
    }

    try:
        _result = module.ipa_command("host_show", name, _args)
    except ipalib_errors.NotFound as e:
        msg = str(e)
        if "host not found" in msg:
            return None
        module.fail_json(msg="host_show failed: %s" % msg)

    _res = _result["result"]
    certs = _res.get("usercertificate")
    if certs is not None:
        _res["usercertificate"] = [encode_certificate(cert) for
                                   cert in certs]
    return _res


def find_dnsrecord(module, name):
    """
    Search for a DNS record.

    This function may raise ipalib_errors.NotFound in some cases,
    and it should be handled by the caller.
    """
    domain_name = name[name.find(".") + 1:]
    host_name = name[:name.find(".")]

    _args = {
        "all": True,
        "idnsname": host_name
    }

    _result = module.ipa_command("dnsrecord_show", domain_name, _args)

    return _result["result"]


def show_host(module, name):
    _result = module.ipa_command("host_show", name, {})
    return _result["result"]


def gen_args(description, locality, location, platform, os, password, random,
             mac_address, sshpubkey, userclass, auth_ind, requires_pre_auth,
             ok_as_delegate, ok_to_auth_as_delegate, force, _reverse,
             ip_address, update_dns):
    # certificate, managedby_host, principal, create_keytab_* and
    # allow_retrieve_keytab_* are not handled here
    _args = {}
    if description is not None:
        _args["description"] = description
    if locality is not None:
        _args["l"] = locality
    if location is not None:
        _args["nshostlocation"] = location
    if platform is not None:
        _args["nshardwareplatform"] = platform
    if os is not None:
        _args["nsosversion"] = os
    if password is not None:
        _args["userpassword"] = password
    if random is not None:
        _args["random"] = random
    if mac_address is not None:
        _args["macaddress"] = mac_address
    if sshpubkey is not None:
        _args["ipasshpubkey"] = sshpubkey
    if userclass is not None:
        _args["userclass"] = userclass
    if auth_ind is not None:
        _args["krbprincipalauthind"] = auth_ind
    if requires_pre_auth is not None:
        _args["ipakrbrequirespreauth"] = requires_pre_auth
    if ok_as_delegate is not None:
        _args["ipakrbokasdelegate"] = ok_as_delegate
    if ok_to_auth_as_delegate is not None:
        _args["ipakrboktoauthasdelegate"] = ok_to_auth_as_delegate
    if force is not None:
        _args["force"] = force
    if ip_address is not None:
        # IP addresses are handed extra, therefore it is needed to set
        # the force option here to make sure that host-add is able to
        # add a host without IP address.
        _args["force"] = True
    if update_dns is not None:
        _args["updatedns"] = update_dns

    return _args


def gen_dnsrecord_args(module, ip_address, reverse):
    _args = {}
    if reverse is not None:
        _args["a_extra_create_reverse"] = reverse
        _args["aaaa_extra_create_reverse"] = reverse
    if ip_address is not None:
        for ip in ip_address:
            if is_ipv4_addr(ip):
                _args.setdefault("arecord", []).append(ip)
            elif is_ipv6_addr(ip):
                _args.setdefault("aaaarecord", []).append(ip)
            else:
                module.fail_json(msg="'%s' is not a valid IP address." % ip)

    return _args


def check_parameters(   # pylint: disable=unused-argument
        module, state, action,
        description, locality, location, platform, os, password, random,
        certificate, managedby_host, principal, allow_create_keytab_user,
        allow_create_keytab_group, allow_create_keytab_host,
        allow_create_keytab_hostgroup, allow_retrieve_keytab_user,
        allow_retrieve_keytab_group, allow_retrieve_keytab_host,
        allow_retrieve_keytab_hostgroup, mac_address, sshpubkey,
        userclass, auth_ind, requires_pre_auth, ok_as_delegate,
        ok_to_auth_as_delegate, force, reverse, ip_address, update_dns,
        update_password):
    invalid = []
    if state == "present":
        if action == "member":
            # certificate, managedby_host, principal,
            # allow_create_keytab_*, allow_retrieve_keytab_*,
            invalid = ["description", "locality", "location", "platform",
                       "os", "password", "random", "mac_address", "sshpubkey",
                       "userclass", "auth_ind", "requires_pre_auth",
                       "ok_as_delegate", "ok_to_auth_as_delegate", "force",
                       "reverse", "update_dns", "update_password"]

    if state == "absent":
        invalid = ["description", "locality", "location", "platform", "os",
                   "password", "random", "mac_address", "sshpubkey",
                   "userclass", "auth_ind", "requires_pre_auth",
                   "ok_as_delegate", "ok_to_auth_as_delegate", "force",
                   "reverse", "update_password"]
        if action == "host":
            invalid = [
                "certificate", "managedby_host", "principal",
                "allow_create_keytab_user", "allow_create_keytab_group",
                "allow_create_keytab_host", "allow_create_keytab_hostgroup",
                "allow_retrieve_keytab_user", "allow_retrieve_keytab_group",
                "allow_retrieve_keytab_host",
                "allow_retrieve_keytab_hostgroup"
            ]

    module.params_fail_used_invalid(invalid, state, action)


# pylint: disable=unused-argument
def result_handler(module, result, command, name, args, errors, exit_args,
                   one_name):
    if "random" in args and command in ["host_add", "host_mod"] \
       and "randompassword" in result["result"]:
        if one_name:
            exit_args["randompassword"] = \
                result["result"]["randompassword"]
        else:
            exit_args.setdefault(name, {})["randompassword"] = \
                result["result"]["randompassword"]

    # All "already a member" and "not a member" failures in the
    # result are ignored. All others are reported.
    if "failed" in result and len(result["failed"]) > 0:
        for item in result["failed"]:
            failed_item = result["failed"][item]
            for member_type in failed_item:
                for member, failure in failed_item[member_type]:
                    if "already a member" in failure \
                       or "not a member" in failure:
                        continue
                    errors.append("%s: %s %s: %s" % (
                        command, member_type, member, failure))


# pylint: disable=unused-argument
def exception_handler(module, ex, errors, exit_args, one_name):
    msg = str(ex)
    if "already contains" in msg \
       or "does not contain" in msg:
        return True

    #  The canonical principal name may not be removed
    if "equal to the canonical principal name must" in msg:
        return True

    # Host is already disabled, ignore error
    if "This entry is already disabled" in msg:
        return True

    # Ignore no modification error.
    if "no modifications to be performed" in msg:
        return True

    return False


def main():
    host_spec = dict(
        # present
        description=dict(type="str", default=None),
        locality=dict(type="str", default=None),
        location=dict(type="str", aliases=["ns_host_location"],
                      default=None),
        platform=dict(type="str", aliases=["ns_hardware_platform"],
                      default=None),
        os=dict(type="str", aliases=["ns_os_version"], default=None),
        password=dict(type="str",
                      aliases=["user_password", "userpassword"],
                      default=None, no_log=True),
        random=dict(type="bool", aliases=["random_password"],
                    default=None),
        certificate=dict(type="list", elements="str",
                         aliases=["usercertificate"], default=None),
        managedby_host=dict(type="list", elements="str", default=None),
        principal=dict(type="list", elements="str",
                       aliases=["principalname", "krbprincipalname"],
                       default=None),
        allow_create_keytab_user=dict(
            type="list", elements="str",
            aliases=["ipaallowedtoperform_write_keys_user"],
            default=None, no_log=False),
        allow_create_keytab_group=dict(
            type="list", elements="str",
            aliases=["ipaallowedtoperform_write_keys_group"],
            default=None, no_log=False),
        allow_create_keytab_host=dict(
            type="list", elements="str",
            aliases=["ipaallowedtoperform_write_keys_host"],
            default=None, no_log=False),
        allow_create_keytab_hostgroup=dict(
            type="list", elements="str",
            aliases=["ipaallowedtoperform_write_keys_hostgroup"],
            default=None, no_log=False),
        allow_retrieve_keytab_user=dict(
            type="list", elements="str",
            aliases=["ipaallowedtoperform_read_keys_user"],
            default=None, no_log=False),
        allow_retrieve_keytab_group=dict(
            type="list", elements="str",
            aliases=["ipaallowedtoperform_read_keys_group"],
            default=None, no_log=False),
        allow_retrieve_keytab_host=dict(
            type="list", elements="str",
            aliases=["ipaallowedtoperform_read_keys_host"],
            default=None, no_log=False),
        allow_retrieve_keytab_hostgroup=dict(
            type="list", elements="str",
            aliases=["ipaallowedtoperform_read_keys_hostgroup"],
            default=None, no_log=False),
        mac_address=dict(type="list", elements="str", aliases=["macaddress"],
                         default=None),
        sshpubkey=dict(type="list", elements="str", aliases=["ipasshpubkey"],
                       default=None),
        userclass=dict(type="list", elements="str", aliases=["class"],
                       default=None),
        auth_ind=dict(type='list', elements="str",
                      aliases=["krbprincipalauthind"], default=None,
                      choices=['radius', 'otp', 'pkinit', 'hardened', '']),
        requires_pre_auth=dict(type="bool", aliases=["ipakrbrequirespreauth"],
                               default=None),
        ok_as_delegate=dict(type="bool", aliases=["ipakrbokasdelegate"],
                            default=None),
        ok_to_auth_as_delegate=dict(type="bool",
                                    aliases=["ipakrboktoauthasdelegate"],
                                    default=None),
        force=dict(type='bool', default=None),
        reverse=dict(type='bool', default=None),
        ip_address=dict(type="list", elements="str", aliases=["ipaddress"],
                        default=None),
        update_dns=dict(type="bool", aliases=["updatedns"],
                        default=None),
        # no_members

        # for update:
        # krbprincipalname
    )

    ansible_module = IPAAnsibleModule(
        argument_spec=dict(
            # general
            name=dict(type="list", elements="str", aliases=["fqdn"],
                      default=None, required=False),

            hosts=dict(type="list", default=None,
                       options=dict(
                           # Here name is a simple string
                           name=dict(type="str", aliases=["fqdn"],
                                     required=True),
                           # Add host specific parameters
                           **host_spec
                       ),
                       elements='dict', required=False),

            # mod
            update_password=dict(type='str', default=None, no_log=False,
                                 choices=['always', 'on_create']),

            # general
            action=dict(type="str", default="host",
                        choices=["member", "host"]),
            state=dict(type="str", default="present",
                       choices=["present", "absent", "disabled"]),

            # Add host specific parameters for simple use case
            **host_spec
        ),
        mutually_exclusive=[["name", "hosts"]],
        required_one_of=[["name", "hosts"]],
        supports_check_mode=True,
    )

    ansible_module._ansible_debug = True

    # Get parameters

    # general
    names = ansible_module.params_get("name")
    hosts = ansible_module.params_get("hosts")

    # present
    description = ansible_module.params_get("description")
    locality = ansible_module.params_get("locality")
    location = ansible_module.params_get("location")
    platform = ansible_module.params_get("platform")
    os = ansible_module.params_get("os")
    password = ansible_module.params_get("password")
    random = ansible_module.params_get("random")
    certificate = ansible_module.params_get("certificate")
    managedby_host = ansible_module.params_get("managedby_host")
    principal = ansible_module.params_get("principal")
    allow_create_keytab_user = ansible_module.params_get(
        "allow_create_keytab_user")
    allow_create_keytab_group = ansible_module.params_get(
        "allow_create_keytab_group")
    allow_create_keytab_host = ansible_module.params_get(
        "allow_create_keytab_host")
    allow_create_keytab_hostgroup = ansible_module.params_get(
        "allow_create_keytab_hostgroup")
    allow_retrieve_keytab_user = ansible_module.params_get(
        "allow_retrieve_keytab_user")
    allow_retrieve_keytab_group = ansible_module.params_get(
        "allow_retrieve_keytab_group")
    allow_retrieve_keytab_host = ansible_module.params_get(
        "allow_retrieve_keytab_host")
    allow_retrieve_keytab_hostgroup = ansible_module.params_get(
        "allow_retrieve_keytab_hostgroup")
    mac_address = ansible_module.params_get("mac_address")
    sshpubkey = ansible_module.params_get("sshpubkey",
                                          allow_empty_string=True)
    userclass = ansible_module.params_get("userclass")
    auth_ind = ansible_module.params_get("auth_ind", allow_empty_string=True)
    requires_pre_auth = ansible_module.params_get("requires_pre_auth")
    ok_as_delegate = ansible_module.params_get("ok_as_delegate")
    ok_to_auth_as_delegate = ansible_module.params_get(
        "ok_to_auth_as_delegate")
    force = ansible_module.params_get("force")
    reverse = ansible_module.params_get("reverse")
    ip_address = ansible_module.params_get("ip_address")
    update_dns = ansible_module.params_get("update_dns")
    update_password = ansible_module.params_get("update_password")
    # general
    action = ansible_module.params_get("action")
    state = ansible_module.params_get("state")

    # Check parameters

    if (names is None or len(names) < 1) and \
       (hosts is None or len(hosts) < 1):
        ansible_module.fail_json(msg="One of name and hosts is required")

    if state == "present":
        if names is not None and len(names) != 1:
            ansible_module.fail_json(
                msg="Only one host can be added at a time.")

    check_parameters(
        ansible_module, state, action,
        description, locality, location, platform, os, password, random,
        certificate, managedby_host, principal, allow_create_keytab_user,
        allow_create_keytab_group, allow_create_keytab_host,
        allow_create_keytab_hostgroup, allow_retrieve_keytab_user,
        allow_retrieve_keytab_group, allow_retrieve_keytab_host,
        allow_retrieve_keytab_hostgroup, mac_address, sshpubkey, userclass,
        auth_ind, requires_pre_auth, ok_as_delegate, ok_to_auth_as_delegate,
        force, reverse, ip_address, update_dns, update_password)

    # Use hosts if names is None
    if hosts is not None:
        names = hosts

    # Init

    changed = False
    exit_args = {}

    # Connect to IPA API
    with ansible_module.ipa_connect():

        # Check version specific settings

        server_realm = ansible_module.ipa_get_realm()

        commands = []
        host_set = set()

        for host in names:
            if isinstance(host, dict):
                name = host.get("name")
                if name in host_set:
                    ansible_module.fail_json(
                        msg="host '%s' is used more than once" % name)
                host_set.add(name)
                description = host.get("description")
                locality = host.get("locality")
                location = host.get("location")
                platform = host.get("platform")
                os = host.get("os")
                password = host.get("password")
                random = host.get("random")
                certificate = host.get("certificate")
                managedby_host = host.get("managedby_host")
                principal = host.get("principal")
                allow_create_keytab_user = host.get(
                    "allow_create_keytab_user")
                allow_create_keytab_group = host.get(
                    "allow_create_keytab_group")
                allow_create_keytab_host = host.get(
                    "allow_create_keytab_host")
                allow_create_keytab_hostgroup = host.get(
                    "allow_create_keytab_hostgroup")
                allow_retrieve_keytab_user = host.get(
                    "allow_retrieve_keytab_user")
                allow_retrieve_keytab_group = host.get(
                    "allow_retrieve_keytab_group")
                allow_retrieve_keytab_host = host.get(
                    "allow_retrieve_keytab_host")
                allow_retrieve_keytab_hostgroup = host.get(
                    "allow_retrieve_keytab_hostgroup")
                mac_address = host.get("mac_address")
                sshpubkey = host.get("sshpubkey")
                userclass = host.get("userclass")
                auth_ind = host.get("auth_ind")
                requires_pre_auth = host.get("requires_pre_auth")
                ok_as_delegate = host.get("ok_as_delegate")
                ok_to_auth_as_delegate = host.get("ok_to_auth_as_delegate")
                force = host.get("force")
                reverse = host.get("reverse")
                ip_address = host.get("ip_address")
                update_dns = host.get("update_dns")
                # update_password is not part of hosts structure
                # action is not part of hosts structure
                # state is not part of hosts structure

                check_parameters(
                    ansible_module, state, action,
                    description, locality, location, platform, os, password,
                    random, certificate, managedby_host, principal,
                    allow_create_keytab_user, allow_create_keytab_group,
                    allow_create_keytab_host, allow_create_keytab_hostgroup,
                    allow_retrieve_keytab_user, allow_retrieve_keytab_group,
                    allow_retrieve_keytab_host,
                    allow_retrieve_keytab_hostgroup, mac_address, sshpubkey,
                    userclass, auth_ind, requires_pre_auth, ok_as_delegate,
                    ok_to_auth_as_delegate, force, reverse, ip_address,
                    update_dns, update_password)

            elif isinstance(host, (str, unicode)):
                name = host
            else:
                ansible_module.fail_json(msg="Host '%s' is not valid" %
                                         repr(host))

            # Make sure host exists
            res_find = find_host(ansible_module, name)
            try:
                res_find_dnsrecord = find_dnsrecord(ansible_module, name)
            except ipalib_errors.NotFound as e:
                msg = str(e)
                dns_not_configured = "DNS is not configured" in msg
                dns_zone_not_found = "DNS zone not found" in msg
                dns_res_not_found = "DNS resource record not found" in msg
                if (
                    dns_res_not_found
                    or ip_address is None
                    and (dns_not_configured or dns_zone_not_found)
                ):
                    # IP address(es) not given and no DNS support in IPA
                    # -> Ignore failure
                    # IP address(es) not given and DNS zone is not found
                    # -> Ignore failure
                    res_find_dnsrecord = None
                else:
                    ansible_module.fail_json(msg="%s: %s" % (host, msg))

            # Create command
            if state == "present":
                # Generate args
                args = gen_args(
                    description, locality, location, platform, os, password,
                    random, mac_address, sshpubkey, userclass, auth_ind,
                    requires_pre_auth, ok_as_delegate, ok_to_auth_as_delegate,
                    force, reverse, ip_address, update_dns)
                dnsrecord_args = gen_dnsrecord_args(
                    ansible_module, ip_address, reverse)

                if action == "host":
                    # Found the host
                    if res_find is not None:
                        # Ignore password with update_password == on_create
                        if update_password == "on_create":
                            # Ignore userpassword and random for existing
                            # host if update_password is "on_create"
                            if "userpassword" in args:
                                del args["userpassword"]
                            if "random" in args:
                                del args["random"]
                        elif "userpassword" in args or "random" in args:
                            # Allow an existing OTP to be reset but don't
                            # allow a OTP or to be added to an enrolled host.
                            # Also do not allow to change the password for an
                            # enrolled host.

                            if not res_find["has_password"] and \
                               res_find["has_keytab"]:
                                ansible_module.fail_json(
                                    msg="%s: Password cannot be set on "
                                    "enrolled host." % host
                                )

                        # Ignore force, ip_address and no_reverse for mod
                        for x in ["force", "ip_address", "no_reverse"]:
                            if x in args:
                                del args[x]

                        # Ignore auth_ind if it is empty (for resetting)
                        # and not set in for the host
                        if "krbprincipalauthind" not in res_find and \
                           "krbprincipalauthind" in args and \
                           args["krbprincipalauthind"] == ['']:
                            del args["krbprincipalauthind"]

                        # For all settings is args, check if there are
                        # different settings in the find result.
                        # If yes: modify
                        if not compare_args_ipa(ansible_module, args,
                                                res_find):
                            commands.append([name, "host_mod", args])
                        elif random and "userpassword" in res_find:
                            # Host exists and random is set, return
                            # userpassword
                            if len(names) == 1:
                                exit_args["userpassword"] = \
                                    res_find["userpassword"]
                            else:
                                exit_args.setdefault("hosts", {})[name] = {
                                    "userpassword": res_find["userpassword"]
                                }

                    else:
                        # Remove update_dns as it is not supported by host_add
                        if "updatedns" in args:
                            del args["updatedns"]
                        commands.append([name, "host_add", args])

                    # Handle members: certificate, managedby_host, principal,
                    # allow_create_keytab and allow_retrieve_keytab
                    if res_find is not None:
                        certificate_add, certificate_del = gen_add_del_lists(
                            certificate, res_find.get("usercertificate"))
                        managedby_host_add, managedby_host_del = \
                            gen_add_del_lists(managedby_host,
                                              res_find.get("managedby_host"))
                        principal_add, principal_del = gen_add_del_lists(
                            principal, res_find.get("principal"))
                        # Principals are not returned as utf8 for IPA using
                        # python2 using host_show, therefore we need to
                        # convert the principals that we should remove.
                        principal_del = [unicode(x) for x in principal_del]

                        (allow_create_keytab_user_add,
                         allow_create_keytab_user_del) = \
                            gen_add_del_lists(
                                allow_create_keytab_user,
                                res_find.get(
                                    "ipaallowedtoperform_write_keys_user"))
                        (allow_create_keytab_group_add,
                         allow_create_keytab_group_del) = \
                            gen_add_del_lists(
                                allow_create_keytab_group,
                                res_find.get(
                                    "ipaallowedtoperform_write_keys_group"))
                        (allow_create_keytab_host_add,
                         allow_create_keytab_host_del) = \
                            gen_add_del_lists(
                                allow_create_keytab_host,
                                res_find.get(
                                    "ipaallowedtoperform_write_keys_host"))
                        (allow_create_keytab_hostgroup_add,
                         allow_create_keytab_hostgroup_del) = \
                            gen_add_del_lists(
                                allow_create_keytab_hostgroup,
                                res_find.get(
                                    "ipaallowedtoperform_write_keys_"
                                    "hostgroup"))
                        (allow_retrieve_keytab_user_add,
                         allow_retrieve_keytab_user_del) = \
                            gen_add_del_lists(
                                allow_retrieve_keytab_user,
                                res_find.get(
                                    "ipaallowedtoperform_read_keys_user"))
                        (allow_retrieve_keytab_group_add,
                         allow_retrieve_keytab_group_del) = \
                            gen_add_del_lists(
                                allow_retrieve_keytab_group,
                                res_find.get(
                                    "ipaallowedtoperform_read_keys_group"))
                        (allow_retrieve_keytab_host_add,
                         allow_retrieve_keytab_host_del) = \
                            gen_add_del_lists(
                                allow_retrieve_keytab_host,
                                res_find.get(
                                    "ipaallowedtoperform_read_keys_host"))
                        (allow_retrieve_keytab_hostgroup_add,
                         allow_retrieve_keytab_hostgroup_del) = \
                            gen_add_del_lists(
                                allow_retrieve_keytab_hostgroup,
                                res_find.get(
                                    "ipaallowedtoperform_read_keys_hostgroup"))

                        # IP addresses are not really a member of hosts, but
                        # we will simply treat it as this to enable the
                        # addition and removal of IPv4 and IPv6 addresses in
                        # a simple way.
                        _dnsrec = res_find_dnsrecord or {}
                        dnsrecord_a_add, dnsrecord_a_del = gen_add_del_lists(
                            dnsrecord_args.get("arecord"),
                            _dnsrec.get("arecord"))
                        dnsrecord_aaaa_add, dnsrecord_aaaa_del = \
                            gen_add_del_lists(
                                dnsrecord_args.get("aaaarecord"),
                                _dnsrec.get("aaaarecord"))

                else:
                    if res_find is None:
                        ansible_module.fail_json(
                            msg="No host '%s'" % name)

                if action != "host" or (action == "host" and res_find is None):
                    certificate_add = certificate or []
                    certificate_del = []
                    managedby_host_add = managedby_host or []
                    managedby_host_del = []
                    principal_add = principal or []
                    principal_del = []
                    allow_create_keytab_user_add = \
                        allow_create_keytab_user or []
                    allow_create_keytab_user_del = []
                    allow_create_keytab_group_add = \
                        allow_create_keytab_group or []
                    allow_create_keytab_group_del = []
                    allow_create_keytab_host_add = \
                        allow_create_keytab_host or []
                    allow_create_keytab_host_del = []
                    allow_create_keytab_hostgroup_add = \
                        allow_create_keytab_hostgroup or []
                    allow_create_keytab_hostgroup_del = []
                    allow_retrieve_keytab_user_add = \
                        allow_retrieve_keytab_user or []
                    allow_retrieve_keytab_user_del = []
                    allow_retrieve_keytab_group_add = \
                        allow_retrieve_keytab_group or []
                    allow_retrieve_keytab_group_del = []
                    allow_retrieve_keytab_host_add = \
                        allow_retrieve_keytab_host or []
                    allow_retrieve_keytab_host_del = []
                    allow_retrieve_keytab_hostgroup_add = \
                        allow_retrieve_keytab_hostgroup or []
                    allow_retrieve_keytab_hostgroup_del = []
                    dnsrecord_a_add = dnsrecord_args.get("arecord") or []
                    dnsrecord_a_del = []
                    dnsrecord_aaaa_add = dnsrecord_args.get("aaaarecord") or []
                    dnsrecord_aaaa_del = []

                # Remove canonical principal from principal_del
                canonical_principal = "host/" + name + "@" + server_realm
                if canonical_principal in principal_del and \
                   action == "host" and (principal is not None or
                                         canonical_principal not in principal):
                    principal_del.remove(canonical_principal)

                # Remove canonical managedby managedby_host_del for
                # action host if managedby_host is set and the canonical
                # managedby host is not in the managedby_host list.
                canonical_managedby_host = name
                if canonical_managedby_host in managedby_host_del and \
                   action == "host" and (managedby_host is None or
                                         canonical_managedby_host not in
                                         managedby_host):
                    managedby_host_del.remove(canonical_managedby_host)

                # Certificates need to be added and removed one by one,
                # because if entry already exists, the processing of
                # the remaining enries is stopped. The same applies to
                # the removal of non-existing entries.

                # Add certificates
                for _certificate in certificate_add:
                    commands.append([name, "host_add_cert",
                                     {
                                         "usercertificate":
                                         _certificate,
                                     }])
                # Remove certificates
                for _certificate in certificate_del:
                    commands.append([name, "host_remove_cert",
                                     {
                                         "usercertificate":
                                         _certificate,
                                     }])

                # Managedby_Hosts need to be added and removed one by one,
                # because if entry already exists, the processing of
                # the remaining enries is stopped. The same applies to
                # the removal of non-existing entries.

                # Add managedby_hosts
                for _managedby_host in managedby_host_add:
                    commands.append([name, "host_add_managedby",
                                     {
                                         "host":
                                         _managedby_host,
                                     }])
                # Remove managedby_hosts
                for _managedby_host in managedby_host_del:
                    commands.append([name, "host_remove_managedby",
                                     {
                                         "host":
                                         _managedby_host,
                                     }])

                # Principals need to be added and removed one by one,
                # because if entry already exists, the processing of
                # the remaining enries is stopped. The same applies to
                # the removal of non-existing entries.

                # Add principals
                for _principal in principal_add:
                    commands.append([name, "host_add_principal",
                                     {
                                         "krbprincipalname":
                                         _principal,
                                     }])
                # Remove principals
                for _principal in principal_del:
                    commands.append([name, "host_remove_principal",
                                     {
                                         "krbprincipalname":
                                         _principal,
                                     }])

                # Allow create keytab
                if len(allow_create_keytab_user_add) > 0 or \
                   len(allow_create_keytab_group_add) > 0 or \
                   len(allow_create_keytab_host_add) > 0 or \
                   len(allow_create_keytab_hostgroup_add) > 0:
                    commands.append(
                        [name, "host_allow_create_keytab",
                         {
                             "user": allow_create_keytab_user_add,
                             "group": allow_create_keytab_group_add,
                             "host": allow_create_keytab_host_add,
                             "hostgroup": allow_create_keytab_hostgroup_add,
                         }])

                # Disallow create keytab
                if len(allow_create_keytab_user_del) > 0 or \
                   len(allow_create_keytab_group_del) > 0 or \
                   len(allow_create_keytab_host_del) > 0 or \
                   len(allow_create_keytab_hostgroup_del) > 0:
                    commands.append(
                        [name, "host_disallow_create_keytab",
                         {
                             "user": allow_create_keytab_user_del,
                             "group": allow_create_keytab_group_del,
                             "host": allow_create_keytab_host_del,
                             "hostgroup": allow_create_keytab_hostgroup_del,
                         }])

                # Allow retrieve keytab
                if len(allow_retrieve_keytab_user_add) > 0 or \
                   len(allow_retrieve_keytab_group_add) > 0 or \
                   len(allow_retrieve_keytab_host_add) > 0 or \
                   len(allow_retrieve_keytab_hostgroup_add) > 0:
                    commands.append(
                        [name, "host_allow_retrieve_keytab",
                         {
                             "user": allow_retrieve_keytab_user_add,
                             "group": allow_retrieve_keytab_group_add,
                             "host": allow_retrieve_keytab_host_add,
                             "hostgroup": allow_retrieve_keytab_hostgroup_add,
                         }])

                # Disallow retrieve keytab
                if len(allow_retrieve_keytab_user_del) > 0 or \
                   len(allow_retrieve_keytab_group_del) > 0 or \
                   len(allow_retrieve_keytab_host_del) > 0 or \
                   len(allow_retrieve_keytab_hostgroup_del) > 0:
                    commands.append(
                        [name, "host_disallow_retrieve_keytab",
                         {
                             "user": allow_retrieve_keytab_user_del,
                             "group": allow_retrieve_keytab_group_del,
                             "host": allow_retrieve_keytab_host_del,
                             "hostgroup": allow_retrieve_keytab_hostgroup_del,
                         }])

                if len(dnsrecord_a_add) > 0 or len(dnsrecord_aaaa_add) > 0:
                    domain_name = name[name.find(".") + 1:]
                    host_name = name[:name.find(".")]

                    _args = {"idnsname": host_name}
                    if len(dnsrecord_a_add) > 0:
                        _args["arecord"] = dnsrecord_a_add
                        if reverse is not None:
                            _args["a_extra_create_reverse"] = reverse
                    if len(dnsrecord_aaaa_add) > 0:
                        _args["aaaarecord"] = dnsrecord_aaaa_add
                        if reverse is not None:
                            _args["aaaa_extra_create_reverse"] = reverse

                    commands.append([domain_name,
                                     "dnsrecord_add", _args])

                if len(dnsrecord_a_del) > 0 or len(dnsrecord_aaaa_del) > 0:
                    domain_name = name[name.find(".") + 1:]
                    host_name = name[:name.find(".")]

                    # There seems to be an issue with dnsrecord_del (not
                    # for dnsrecord_add) if aaaarecord is an empty list.
                    # Therefore this is done differently here:
                    _args = {"idnsname": host_name}
                    if len(dnsrecord_a_del) > 0:
                        _args["arecord"] = dnsrecord_a_del
                    if len(dnsrecord_aaaa_del) > 0:
                        _args["aaaarecord"] = dnsrecord_aaaa_del

                    commands.append([domain_name,
                                     "dnsrecord_del", _args])

            elif state == "absent":
                if action == "host":

                    if res_find is not None:
                        args = {}
                        if update_dns is not None:
                            args["updatedns"] = update_dns
                        commands.append([name, "host_del", args])
                else:

                    # Certificates need to be added and removed one by one,
                    # because if entry already exists, the processing of
                    # the remaining enries is stopped. The same applies to
                    # the removal of non-existing entries.

                    # Remove certificates
                    if certificate is not None:
                        for _certificate in certificate:
                            commands.append([name, "host_remove_cert",
                                             {
                                                 "usercertificate":
                                                 _certificate,
                                             }])

                    # Managedby_Hosts need to be added and removed one by one,
                    # because if entry already exists, the processing of
                    # the remaining enries is stopped. The same applies to
                    # the removal of non-existing entries.

                    # Remove managedby_hosts
                    if managedby_host is not None:
                        for _managedby_host in managedby_host:
                            commands.append([name, "host_remove_managedby",
                                             {
                                                 "host":
                                                 _managedby_host,
                                             }])

                    # Principals need to be added and removed one by one,
                    # because if entry already exists, the processing of
                    # the remaining enries is stopped. The same applies to
                    # the removal of non-existing entries.

                    # Remove principals
                    if principal is not None:
                        for _principal in principal:
                            commands.append([name, "host_remove_principal",
                                             {
                                                 "krbprincipalname":
                                                 _principal,
                                             }])

                    # Disallow create keytab
                    if allow_create_keytab_user is not None or \
                       allow_create_keytab_group is not None or \
                       allow_create_keytab_host is not None or \
                       allow_create_keytab_hostgroup is not None:
                        commands.append(
                            [name, "host_disallow_create_keytab",
                             {
                                 "user": allow_create_keytab_user,
                                 "group": allow_create_keytab_group,
                                 "host": allow_create_keytab_host,
                                 "hostgroup": allow_create_keytab_hostgroup,
                             }])

                    # Disallow retrieve keytab
                    if allow_retrieve_keytab_user is not None or \
                       allow_retrieve_keytab_group is not None or \
                       allow_retrieve_keytab_host is not None or \
                       allow_retrieve_keytab_hostgroup is not None:
                        commands.append(
                            [name, "host_disallow_retrieve_keytab",
                             {
                                 "user": allow_retrieve_keytab_user,
                                 "group": allow_retrieve_keytab_group,
                                 "host": allow_retrieve_keytab_host,
                                 "hostgroup": allow_retrieve_keytab_hostgroup,
                             }])

                    dnsrecord_args = gen_dnsrecord_args(ansible_module,
                                                        ip_address, reverse)

                    # Remove arecord and aaaarecord from dnsrecord_args
                    # if the record does not exits in res_find_dnsrecord
                    # to prevent "DNS resource record not found" error
                    if "arecord" in dnsrecord_args \
                       and dnsrecord_args["arecord"] is not None \
                       and len(dnsrecord_args["arecord"]) > 0 \
                       and (res_find_dnsrecord is None
                            or "arecord" not in res_find_dnsrecord):
                        del dnsrecord_args["arecord"]
                    if "aaaarecord" in dnsrecord_args \
                       and dnsrecord_args["aaaarecord"] is not None \
                       and len(dnsrecord_args["aaaarecord"]) > 0 \
                       and (res_find_dnsrecord is None
                            or "aaaarecord" not in res_find_dnsrecord):
                        del dnsrecord_args["aaaarecord"]

                    if "arecord" in dnsrecord_args or \
                       "aaaarecord" in dnsrecord_args:
                        domain_name = name[name.find(".") + 1:]
                        host_name = name[:name.find(".")]
                        dnsrecord_args["idnsname"] = host_name

                        commands.append([domain_name, "dnsrecord_del",
                                         dnsrecord_args])

            elif state == "disabled":
                if res_find is not None:
                    commands.append([name, "host_disable", {}])
                else:
                    raise ValueError("No host '%s'" % name)

            else:
                ansible_module.fail_json(msg="Unkown state '%s'" % state)

        del host_set

        # Execute commands

        changed = ansible_module.execute_ipa_commands(
            commands, result_handler, exception_handler,
            exit_args=exit_args, one_name=len(names) == 1)

    # Done

    ansible_module.exit_json(changed=changed, host=exit_args)


if __name__ == "__main__":
    main()



# Authors:
#   Sergio Oliveira Campos <seocam@redhat.com>
#
# Copyright (C) 2020 Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from pytest_sourceorder import ordered

from utils import AnsibleFreeIPATestCase, kinit_admin, kdestroy

BASE_PATH = "pytests/dnszone/playbooks/"


@ordered
class TestDNSZone(AnsibleFreeIPATestCase):
    def test_dnszone_add_without_forwarder(self):
        """TC-01: Add dns zone without forwarder."""
        zone01 = "01testzone.test"
        self.check_notexists([zone01], "dnszone-find")
        self.run_playbook(BASE_PATH + "dnszone_add_without_forwarder.yaml")
        self.check_details([zone01], "dnszone-find")

    def test_dnszone_add_multiple_ipv4_ipv6_forwarders(self):
        """TC-04: Update multiple ipv4 and ipv6 forwarders."""
        zone04 = "04testzone.test"
        self.check_notexists([zone04], "dnszone-find")

        # add dns zone with multiple forwarders
        self.run_playbook(
            (BASE_PATH + "dnszone_add_multiple_ipv4_ipv6_forwarders.yaml")
        )

        exp_forwarders = [
            "192.11.22.33",
            "192.11.22.34 port 23",
            "2001:db8:cafe:1::1",
            "2001:db8:cafe:1::4 port 34",
        ]
        exp_forwarders = ", ".join(exp_forwarders)
        self.check_details([exp_forwarders], "dnszone-find", [zone04])

    def test_dnszone_with_forward_policy_only(self):
        """TC-26: Add DNS zone with forward_policy only."""
        zone26 = "26testzone.test"
        self.check_notexists([zone26], "dnszone-find")
        # add dns zone
        self.run_playbook(BASE_PATH + "dnszone_with_forward_policy_only.yaml")
        self.check_details(["Forward policy: only"], "dnszone-find", [zone26])

    def test_dnszone_disable(self):
        """TC-30: Disable DNS Zone."""
        zone26 = "26testzone.test"
        self.check_details(
            ["Active zone: (TRUE|True)"], "dnszone-find", [zone26]
        )
        # Disable dns zone
        self.run_playbook(BASE_PATH + "dnszone_disable.yaml")
        self.check_details(
            ["Active zone: (FALSE|False)"], "dnszone-find", [zone26]
        )

    def test_dnszone_enable(self):
        """TC-31: Enable DNS Zone."""
        zone26 = "26testzone.test"
        self.check_details(
            ["Active zone: (FALSE|False)"], "dnszone-find", [zone26]
        )
        # Enable dns zone
        self.run_playbook(BASE_PATH + "dnszone_enable.yaml")
        self.check_details(
            ["Active zone: (TRUE|True)"], "dnszone-find", [zone26]
        )

    def test_dnszone_name_from_ip(self):
        """TC-35: Add dns zone with reverse zone IP. Bug#1845056."""
        zone = "8.192.in-addr.arpa."
        expected_msg = "Zone name: {0}".format(zone)
        self.check_notexists([expected_msg], "dnszone-find", [zone])

        self.mark_xfail_using_ansible_freeipa_version(
            version="ansible-freeipa-0.1.12-5.el8.noarch",
            reason="Fix is not available for BZ-1845056",
        )

        self.run_playbook(BASE_PATH + "dnszone_name_from_ip.yaml")
        self.check_details([expected_msg], "dnszone-find", [zone])

    def test_dnszone_del_multiple(self):
        """TC-33: Delete multiple DNS zones Bug#1845058."""
        zone = ["delzone1.com", "delzone2.com", "delzone3.com"]
        for add_zone in zone:
            kinit_admin(self.master)
            self.master.run("ipa dnszone-add " + add_zone)
            self.check_details([add_zone], "dnszone-show", [add_zone])
            kdestroy(self.master)

        self.mark_xfail_using_ansible_freeipa_version(
            version="ansible-freeipa-0.1.12-5.el8.noarch",
            reason="Fix is not available for BZ-1845058",
        )

        self.run_playbook(BASE_PATH + "dnszone_del_multiple.yaml")
        # verify multiple dnszones are removed
        for add_zone in zone:
            error = "ipa: ERROR: {0}.: DNS zone not found".format(add_zone)
            self.check_notexists([error], "dnszone-show", [add_zone])

    def test_dnszone_invalid_ip(self):
        """TC-07: Update with invalid IPs in allow_transfer. Bug#1845051."""
        invalid_zone_name = "invalidzone.test"
        invalid_zone_ip = "in.va.li.d"
        expected_error = "Invalid IP for DNS forwarder"

        self.mark_xfail_using_ansible_freeipa_version(
            version="ansible-freeipa-0.1.12-5.el8.noarch",
            reason="Fix is not available for BZ-1845058",
        )

        self.run_playbook_with_exp_msg(
            BASE_PATH + "dnszone_invalid_ip.yaml", expected_error,
        )
        self.check_notexists(
            [invalid_zone_ip], "dnszone-show", [invalid_zone_name],
        )



# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Based on ipa-client-install code
#
# Copyright (C) 2017-2022  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.0',
    'supported_by': 'community',
    'status': ['preview'],
}

DOCUMENTATION = '''
---
module: ipaclient_test
short_description: Tries to discover IPA server
description:
  Tries to discover IPA server using DNS or host name
options:
  domain:
    description: Primary DNS domain of the IPA deployment
    type: str
    required: no
  servers:
    description: Fully qualified name of IPA servers to enroll to
    type: list
    elements: str
    required: no
  realm:
    description: Kerberos realm name of the IPA deployment
    type: str
    required: no
  hostname:
    description: Fully qualified name of this host
    type: str
    required: no
  ntp_servers:
    description: ntp servers to use
    type: list
    elements: str
    required: no
  ntp_pool:
    description: ntp server pool to use
    type: str
    required: no
  no_ntp:
    description: Do not configure ntp
    type: bool
    required: no
    default: no
  force_ntpd:
    description:
      Stop and disable any time&date synchronization services besides ntpd
      Deprecated since 4.7
    type: bool
    required: no
    default: no
  nisdomain:
    description: The NIS domain name
    type: str
    required: no
  no_nisdomain:
    description: Do not configure NIS domain name
    type: bool
    required: no
    default: no
  kinit_attempts:
    description: Repeat the request for host Kerberos ticket X times
    type: int
    required: no
  ca_cert_files:
    description:
      List of files containing CA certificates for the service certificate
      files
    type: list
    elements: str
    required: no
  configure_firefox:
    description: Configure Firefox to use IPA domain credentials
    type: bool
    required: no
    default: no
  firefox_dir:
    description:
      Specify directory where Firefox is installed (for example
      '/usr/lib/firefox')
    type: str
    required: no
  ip_addresses:
    description: List of Master Server IP Addresses
    type: list
    elements: str
    required: no
  all_ip_addresses:
    description:
      All routable IP addresses configured on any interface will be added
      to DNS
    type: bool
    required: no
    default: no
  on_master:
    description: Whether the configuration is done on the master or not
    type: bool
    required: no
    default: no
  enable_dns_updates:
    description:
      Configures the machine to attempt dns updates when the ip address
      changes
    type: bool
    required: no
    default: no
author:
    - Thomas Woerner (@t-woerner)
'''

EXAMPLES = '''
# Complete autodiscovery, register return values as ipaclient_test
- name: IPA discovery
  ipaclient_test:
  register: register_ipaclient_test

# Discovery using servers, register return values as ipaclient_test
- name: IPA discovery
  ipaclient_test:
    servers: server1.domain.com,server2.domain.com
  register: register_ipaclient_test

# Discovery using domain name, register return values as ipaclient_test
- name: IPA discovery
  ipaclient_test:
    domain: domain.com
  register: register_ipaclient_test

# Discovery using realm, register return values as ipaclient_test
- name: IPA discovery
  ipaclient_test:
    realm: DOMAIN.COM
  register: register_ipaclient_test

# Discovery using hostname, register return values as ipaclient_test
- name: IPA discovery
  ipaclient_test:
    hostname: host.domain.com
  register: register_ipaclient_test
'''

RETURN = '''
servers:
  description: The list of detected or passed in IPA servers.
  returned: always
  type: list
  elements: str
  sample: ["server1.example.com","server2.example.com"]
domain:
  description: The DNS domain of the detected or passed in IPA deployment.
  returned: always
  type: str
  sample: example.com
realm:
  description: The Kerberos realm of the detected or passed in IPA deployment.
  returned: always
  type: str
  sample: EXAMPLE.COM
kdc:
  description: The detected KDC server name.
  returned: always
  type: str
  sample: server1.example.com
basedn:
  description: The basedn of the detected IPA server.
  returned: always
  type: str
  sample: dc=example,dc=com
hostname:
  description: The detected or passed in FQDN hostname of the client.
  returned: always
  type: str
  sample: client1.example.com
client_domain:
  description: The domain name of the client.
  returned: always
  type: str
  sample: example.com
dnsok:
  description: True if DNS discovery worked and not passed in any servers.
  returned: always
  type: bool
ntp_servers:
  description: The list of detected NTP servers.
  returned: always
  type: list
  elements: str
  sample: ["ntp.example.com"]
ipa_python_version:
  description: >
    The IPA python version as a number:
    <major version>*10000+<minor version>*100+<release>
  returned: always
  type: int
  sample: 040400
nosssd_files:
  description: >
    The dist of nss_ldap or nss-pam-ldapd files if sssd is disabled
  returned: always
  type: list
  elements: str
'''

import os
import socket

try:
    from ansible.module_utils.six.moves.configparser import RawConfigParser
except ImportError:
    from ConfigParser import RawConfigParser

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.ansible_ipa_client import (
    setup_logging, check_imports,
    paths, sysrestore, options, CheckedIPAddress, validate_domain_name,
    logger, x509, normalize_hostname, installer, version, ScriptError,
    CLIENT_INSTALL_ERROR, tasks, check_ldap_conf, timeconf, constants,
    validate_hostname, nssldap_exists, gssapi, remove_file,
    check_ip_addresses, ipadiscovery, print_port_conf_info,
    IPA_PYTHON_VERSION, getargspec
)


def get_cert_path(cert_path):
    """
    If a CA certificate is passed in on the command line, use that.

    Else if a CA file exists in paths.IPA_CA_CRT then use that.

    Otherwise return None.
    """
    if cert_path is not None:
        return cert_path

    if os.path.exists(paths.IPA_CA_CRT):
        return paths.IPA_CA_CRT

    return None


def is_client_configured():
    """
    Check if ipa client is configured.

    IPA client is configured when /etc/ipa/default.conf exists and
    /var/lib/ipa-client/sysrestore/sysrestore.state exists.

    :returns: boolean
    """
    return (os.path.isfile(paths.IPA_DEFAULT_CONF) and
            os.path.isfile(os.path.join(paths.IPA_CLIENT_SYSRESTORE,
                                        sysrestore.SYSRESTORE_STATEFILE)))


def get_ipa_conf():
    """
    Return IPA configuration read from `/etc/ipa/default.conf`.

    :returns: dict containing key,value
    """
    parser = RawConfigParser()
    parser.read(paths.IPA_DEFAULT_CONF)
    result = {}
    for item in ['basedn', 'realm', 'domain', 'server', 'host', 'xmlrpc_uri']:
        if parser.has_option('global', item):
            value = parser.get('global', item)
        else:
            value = None
        if value:
            result[item] = value

    return result


def main():
    module = AnsibleModule(
        argument_spec=dict(
            # basic
            domain=dict(required=False, type='str', default=None),
            servers=dict(required=False, type='list', elements='str',
                         default=None),
            realm=dict(required=False, type='str', default=None),
            hostname=dict(required=False, type='str', default=None),
            ntp_servers=dict(required=False, type='list', elements='str',
                             default=None),
            ntp_pool=dict(required=False, type='str', default=None),
            no_ntp=dict(required=False, type='bool', default=False),
            force_ntpd=dict(required=False, type='bool', default=False),
            nisdomain=dict(required=False, type='str', default=None),
            no_nisdomain=dict(required=False, type='bool', default='no'),
            kinit_attempts=dict(required=False, type='int'),
            ca_cert_files=dict(required=False, type='list', elements='str',
                               default=None),
            configure_firefox=dict(required=False, type='bool', default=False),
            firefox_dir=dict(required=False, type='str'),
            ip_addresses=dict(required=False, type='list', elements='str',
                              default=None),
            all_ip_addresses=dict(required=False, type='bool', default=False),
            on_master=dict(required=False, type='bool', default=False),
            # sssd
            enable_dns_updates=dict(required=False, type='bool',
                                    default=False),
        ),
        supports_check_mode=False,
    )

    # module._ansible_debug = True
    check_imports(module)
    setup_logging()

    options.domain_name = module.params.get('domain')
    options.servers = module.params.get('servers')
    options.realm_name = module.params.get('realm')
    options.host_name = module.params.get('hostname')
    options.ntp_servers = module.params.get('ntp_servers')
    options.ntp_pool = module.params.get('ntp_pool')
    options.no_ntp = module.params.get('no_ntp')
    options.force_ntpd = module.params.get('force_ntpd')
    options.nisdomain = module.params.get('nisdomain')
    options.no_nisdomain = module.params.get('no_nisdomain')
    options.kinit_attempts = module.params.get('kinit_attempts')
    options.ca_cert_files = module.params.get('ca_cert_files')
    options.configure_firefox = module.params.get('configure_firefox')
    options.firefox_dir = module.params.get('firefox_dir')
    options.ip_addresses = module.params.get('ip_addresses')
    options.all_ip_addresses = module.params.get('all_ip_addresses')
    options.on_master = module.params.get('on_master')
    options.enable_dns_updates = module.params.get('enable_dns_updates')

    # Get domain from first server if domain is not set, but if there are
    # servers
    if options.domain_name is None and options.servers is not None:
        if len(options.servers) > 0:
            options.domain_name = options.servers[0][
                options.servers[0].find(".") + 1:]

    try:
        self = options

        # HostNameInstallInterface

        if options.ip_addresses is not None:
            for value in options.ip_addresses:
                try:
                    CheckedIPAddress(value)
                except Exception as e:
                    raise ValueError("invalid IP address {0}: {1}".format(
                        value, e))

        # ServiceInstallInterface

        if options.domain_name:
            validate_domain_name(options.domain_name)

        if options.realm_name:
            # pylint: disable=deprecated-method
            argspec = getargspec(validate_domain_name)
            if "entity" in argspec.args:
                # NUM_VERSION >= 40690:
                validate_domain_name(options.realm_name, entity="realm")

        # ClientInstallInterface

        if options.kinit_attempts < 1:
            raise ValueError("expects an integer greater than 0.")

        # ClientInstallInterface.__init__

        if self.servers and not self.domain_name:
            raise RuntimeError(
                "--server cannot be used without providing --domain")

        if self.force_ntpd:
            logger.warning("Option --force-ntpd has been deprecated")

        if self.ntp_servers and self.no_ntp:
            raise RuntimeError(
                "--ntp-server cannot be used together with --no-ntp")

        if self.ntp_pool and self.no_ntp:
            raise RuntimeError(
                "--ntp-pool cannot be used together with --no-ntp")

        if self.no_nisdomain and self.nisdomain:
            raise RuntimeError(
                "--no-nisdomain cannot be used together with --nisdomain")

        if self.ip_addresses:
            if self.enable_dns_updates:
                raise RuntimeError(
                    "--ip-address cannot be used together with"
                    " --enable-dns-updates")

            if self.all_ip_addresses:
                raise RuntimeError(
                    "--ip-address cannot be used together with"
                    "--all-ip-addresses")

        # SSSDInstallInterface

        self.no_sssd = False

        # ClientInstall

        if options.ca_cert_files is not None:
            for value in options.ca_cert_files:
                if not isinstance(value, list):
                    raise ValueError("Expected list, got {!r}".format(value))
                # this is what init() does
                value = value[-1]
                if not os.path.exists(value):
                    raise ValueError("'%s' does not exist" % value)
                if not os.path.isfile(value):
                    raise ValueError("'%s' is not a file" % value)
                if not os.path.isabs(value):
                    raise ValueError("'%s' is not an absolute file path" %
                                     value)

                try:
                    x509.load_certificate_from_file(value)
                except Exception:
                    raise ValueError("'%s' is not a valid certificate file" %
                                     value)

        # self.prompt_password = self.interactive

        self.no_ac = False

        # ClientInstall.__init__

        if self.firefox_dir and not self.configure_firefox:
            raise RuntimeError(
                "--firefox-dir cannot be used without --configure-firefox "
                "option")

    except (RuntimeError, ValueError) as e:
        module.fail_json(msg=str(e))

    # ipaclient.install.client.init

    # root_logger
    options.debug = False
    if options.domain_name:
        options.domain = normalize_hostname(installer.domain_name)
    else:
        options.domain = None
    options.server = options.servers
    options.realm = options.realm_name
    # installer.primary = installer.fixed_primary
    # if installer.principal:
    #     installer.password = installer.admin_password
    # else:
    #     installer.password = installer.host_password
    installer.hostname = installer.host_name
    options.conf_ntp = not options.no_ntp
    # installer.trust_sshfp = installer.ssh_trust_dns
    # installer.conf_ssh = not installer.no_ssh
    # installer.conf_sshd = not installer.no_sshd
    # installer.conf_sudo = not installer.no_sudo
    # installer.create_sshfp = not installer.no_dns_sshfp
    if installer.ca_cert_files:
        installer.ca_cert_file = installer.ca_cert_files[-1]
    else:
        installer.ca_cert_file = None
    # installer.location = installer.automount_location
    installer.dns_updates = installer.enable_dns_updates
    # installer.krb5_offline_passwords = \
    #     not installer.no_krb5_offline_passwords
    installer.sssd = not installer.no_sssd

    try:

        # client

        # global variables
        hostname = None
        hostname_source = None
        nosssd_files = {}
        dnsok = False
        cli_domain = None
        cli_server = None
        # subject_base = None
        cli_realm = None
        cli_kdc = None
        client_domain = None
        cli_basedn = None
        # end of global variables

        # client.install_check

        logger.info("This program will set up FreeIPA client.")
        logger.info("Version %s", version.VERSION)
        logger.info("")

        cli_domain_source = 'Unknown source'
        cli_server_source = 'Unknown source'

        # fstore = sysrestore.FileStore(paths.IPA_CLIENT_SYSRESTORE)

        if not os.getegid() == 0:
            raise ScriptError(
                "You must be root to run ipa-client-install.",
                rval=CLIENT_INSTALL_ERROR)

        tasks.check_selinux_status()

        # if is_ipa_client_installed(fstore, on_master=options.on_master):
        #     logger.error("IPA client is already configured on this system.")
        #     logger.info(
        #       "If you want to reinstall the IPA client, uninstall it first "
        #       "using 'ipa-client-install --uninstall'.")
        #     raise ScriptError(
        #         "IPA client is already configured on this system.",
        #         rval=CLIENT_ALREADY_CONFIGURED)

        if check_ldap_conf is not None:
            check_ldap_conf()

        if options.conf_ntp:
            try:
                timeconf.check_timedate_services()
            except timeconf.NTPConflictingService as e:
                logger.info(
                    "WARNING: conflicting time&date synchronization service "
                    "'%s' will be disabled in favor of chronyd",
                    e.conflicting_service)
                logger.info("")
            except timeconf.NTPConfigurationError:
                pass

        # password, principal and keytab are checked in tasks/install.yml
        # if options.unattended and (
        #     options.password is None and
        #     options.principal is None and
        #     options.keytab is None and
        #     options.prompt_password is False and
        #     not options.on_master
        # ):
        #     raise ScriptError(
        #         "One of password / principal / keytab is required.",
        #         rval=CLIENT_INSTALL_ERROR)

        if options.hostname:
            hostname = options.hostname
            hostname_source = 'Provided as option'
        else:
            hostname = socket.getfqdn()
            hostname_source = "Machine's FQDN"
        if hostname != hostname.lower():
            raise ScriptError(
                "Invalid hostname '{}', must be lower-case.".format(hostname),
                rval=CLIENT_INSTALL_ERROR
            )

        if hostname in ('localhost', 'localhost.localdomain'):
            raise ScriptError(
                "Invalid hostname, '{}' must not be used.".format(hostname),
                rval=CLIENT_INSTALL_ERROR)

        if hasattr(constants, "MAXHOSTNAMELEN"):
            try:
                validate_hostname(hostname, maxlen=constants.MAXHOSTNAMELEN)
            except ValueError as e:
                raise ScriptError(
                    'invalid hostname: {}'.format(e),
                    rval=CLIENT_INSTALL_ERROR)

        if hasattr(tasks, "is_nosssd_supported"):
            # --no-sssd is not supported any more for rhel-based distros
            if not tasks.is_nosssd_supported() and not options.sssd:
                raise ScriptError(
                    "Option '--no-sssd' is incompatible with the 'authselect' "
                    "tool provided by this distribution for configuring "
                    "system authentication resources",
                    rval=CLIENT_INSTALL_ERROR)

            # --noac is not supported any more for rhel-based distros
            if not tasks.is_nosssd_supported() and options.no_ac:
                raise ScriptError(
                    "Option '--noac' is incompatible with the 'authselect' "
                    "tool provided by this distribution for configuring "
                    "system authentication resources",
                    rval=CLIENT_INSTALL_ERROR)

        # when installing with '--no-sssd' option, check whether nss-ldap is
        # installed
        if not options.sssd:
            if not os.path.exists(paths.PAM_KRB5_SO):
                raise ScriptError(
                    "The pam_krb5 package must be installed",
                    rval=CLIENT_INSTALL_ERROR)

            (nssldap_installed, nosssd_files) = nssldap_exists()
            if not nssldap_installed:
                raise ScriptError(
                    "One of these packages must be installed: nss_ldap or "
                    "nss-pam-ldapd",
                    rval=CLIENT_INSTALL_ERROR)

            # principal and keytab are checked in tasks/install.yml
            # if options.keytab and options.principal:
            #   raise ScriptError(
            #     "Options 'principal' and 'keytab' cannot be used together.",
            #     rval=CLIENT_INSTALL_ERROR)

            # keytab and force_join are checked in tasks/install.yml
            # if options.keytab and options.force_join:
            #   logger.warning("Option 'force-join' has no additional effect "
            #                  "when used with together with option 'keytab'.")

        # Added with freeipa-4.7.1 >>>
        # Remove invalid keytab file
        try:
            gssapi.Credentials(
                store={'keytab': paths.KRB5_KEYTAB},
                usage='accept',
            )
        except gssapi.exceptions.GSSError:
            logger.debug("Deleting invalid keytab: '%s'.", paths.KRB5_KEYTAB)
            remove_file(paths.KRB5_KEYTAB)
        # Added with freeipa-4.7.1 <<<

        # Check if old certificate exist and show warning
        if (
            not options.ca_cert_file and
            get_cert_path(options.ca_cert_file) == paths.IPA_CA_CRT
        ):
            logger.warning("Using existing certificate '%s'.",
                           paths.IPA_CA_CRT)

        if not check_ip_addresses(options):
            raise ScriptError(
                "Failed to check ip addresses, check installation log",
                rval=CLIENT_INSTALL_ERROR)

        # Create the discovery instance
        # pylint: disable=invalid-name
        ds = ipadiscovery.IPADiscovery()

        ret = ds.search(
            domain=options.domain,
            servers=options.server,
            realm=options.realm_name,
            hostname=hostname,
            ca_cert_path=get_cert_path(options.ca_cert_file)
        )

        if options.server and ret != 0:
            # There is no point to continue with installation as server list
            # was passed as a fixed list of server and thus we cannot discover
            # any better result
            logger.error(
                "Failed to verify that %s is an IPA Server.",
                ', '.join(options.server))
            logger.error(
                "This may mean that the remote server is not up "
                "or is not reachable due to network or firewall settings.")
            print_port_conf_info()
            raise ScriptError("Failed to verify that %s is an IPA Server." %
                              ', '.join(options.server),
                              rval=CLIENT_INSTALL_ERROR)

        if ret == ipadiscovery.BAD_HOST_CONFIG:
            logger.error("Can't get the fully qualified name of this host")
            logger.info("Check that the client is properly configured")
            raise ScriptError(
                "Can't get the fully qualified name of this host",
                rval=CLIENT_INSTALL_ERROR)
        if ret == ipadiscovery.NOT_FQDN:
            raise ScriptError(
                "{} is not a fully-qualified hostname".format(hostname),
                rval=CLIENT_INSTALL_ERROR)
        if ret in (ipadiscovery.NO_LDAP_SERVER, ipadiscovery.NOT_IPA_SERVER) \
                or not ds.domain:
            if ret == ipadiscovery.NO_LDAP_SERVER:
                if ds.server:
                    logger.debug("%s is not an LDAP server", ds.server)
                else:
                    logger.debug("No LDAP server found")
            elif ret == ipadiscovery.NOT_IPA_SERVER:
                if ds.server:
                    logger.debug("%s is not an IPA server", ds.server)
                else:
                    logger.debug("No IPA server found")
            else:
                logger.debug("Domain not found")
            if options.domain:
                cli_domain = options.domain
                cli_domain_source = 'Provided as option'
            elif options.unattended:
                raise ScriptError(
                    "Unable to discover domain, not provided on command line",
                    rval=CLIENT_INSTALL_ERROR)
            else:
                raise ScriptError("No interactive installation")
            #    logger.info(
            #        "DNS discovery failed to determine your DNS domain")
            #    cli_domain = user_input(
            #        "Provide the domain name of your IPA server "
            #        "(ex: example.com)",
            #        allow_empty=False)
            #    cli_domain_source = 'Provided interactively'
            #    logger.debug(
            #        "will use interactively provided domain: %s", cli_domain)
            ret = ds.search(
                domain=cli_domain,
                servers=options.server,
                hostname=hostname,
                ca_cert_path=get_cert_path(options.ca_cert_file))

        if not cli_domain:
            if ds.domain:
                cli_domain = ds.domain
                cli_domain_source = ds.domain_source
                logger.debug("will use discovered domain: %s", cli_domain)

        client_domain = hostname[hostname.find(".") + 1:]

        if ret in (ipadiscovery.NO_LDAP_SERVER, ipadiscovery.NOT_IPA_SERVER) \
                or not ds.server:
            logger.debug("IPA Server not found")
            if options.server:
                cli_server = options.server
                cli_server_source = 'Provided as option'
            elif options.unattended:
                raise ScriptError(
                    "Unable to find IPA Server to join",
                    rval=CLIENT_INSTALL_ERROR)
            else:
                raise ScriptError("No interactive installation")
            #    logger.debug("DNS discovery failed to find the IPA Server")
            #    cli_server = [
            #        user_input(
            #            "Provide your IPA server name (ex: ipa.example.com)",
            #            allow_empty=False)
            #    ]
            #    cli_server_source = 'Provided interactively'
            #    logger.debug(
            #      "will use interactively provided server: %s", cli_server[0])
            ret = ds.search(
                domain=cli_domain,
                servers=cli_server,
                hostname=hostname,
                ca_cert_path=get_cert_path(options.ca_cert_file))

        else:
            # Only set dnsok to True if we were not passed in one or more
            # servers and if DNS discovery actually worked.
            if not options.server:
                (server, domain) = ds.check_domain(
                    ds.domain, set(), "Validating DNS Discovery")
                if server and domain:
                    logger.debug("DNS validated, enabling discovery")
                    dnsok = True
                else:
                    logger.debug("DNS discovery failed, disabling discovery")
            else:
                logger.debug(
                    "Using servers from command line, disabling DNS discovery")

        if not cli_server:
            if options.server:
                cli_server = ds.servers
                cli_server_source = 'Provided as option'
                logger.debug(
                    "will use provided server: %s", ', '.join(options.server))
            elif ds.server:
                cli_server = ds.servers
                cli_server_source = ds.server_source
                logger.debug("will use discovered server: %s", cli_server[0])

        if ret == ipadiscovery.NOT_IPA_SERVER:
            logger.error("%s is not an IPA v2 Server.", cli_server[0])
            print_port_conf_info()
            logger.debug("(%s: %s)", cli_server[0], cli_server_source)
            raise ScriptError("%s is not an IPA v2 Server." % cli_server[0],
                              rval=CLIENT_INSTALL_ERROR)

        if ret == ipadiscovery.NO_ACCESS_TO_LDAP:
            logger.warning("Anonymous access to the LDAP server is disabled.")
            logger.info("Proceeding without strict verification.")
            logger.info(
                "Note: This is not an error if anonymous access "
                "has been explicitly restricted.")
            ret = 0

        if ret == ipadiscovery.NO_TLS_LDAP:
            logger.warning(
                "The LDAP server requires TLS is but we do not have the CA.")
            logger.info("Proceeding without strict verification.")
            ret = 0

        if ret != 0:
            logger.error(
                "Failed to verify that %s is an IPA Server.",
                cli_server[0])
            logger.error(
                "This may mean that the remote server is not up "
                "or is not reachable due to network or firewall settings.")
            print_port_conf_info()
            logger.debug("(%s: %s)", cli_server[0], cli_server_source)
            raise ScriptError("Failed to verify that %s is an IPA Server." %
                              cli_server[0],
                              rval=CLIENT_INSTALL_ERROR)

        cli_kdc = ds.kdc
        if dnsok and not cli_kdc:
            logger.error(
                "DNS domain '%s' is not configured for automatic "
                "KDC address lookup.", ds.realm.lower())
            logger.debug("(%s: %s)", ds.realm, ds.realm_source)
            logger.error("KDC address will be set to fixed value.")

        if dnsok:
            logger.info("Discovery was successful!")
        elif not options.unattended:
            raise ScriptError("No interactive installation")
        # if not options.server:
        #     logger.warning(
        #       "The failure to use DNS to find your IPA "
        #       "server indicates that your resolv.conf file is not properly "
        #       "configured.")
        # logger.info(
        #     "Autodiscovery of servers for failover cannot work "
        #     "with this configuration.")
        # logger.info(
        #   "If you proceed with the installation, services "
        #   "will be configured to always access the discovered server for "
        #   "all operations and will not fail over to other servers in case "
        #   "of failure.")
        # if not user_input(
        #     "Proceed with fixed values and no DNS discovery?", False):
        #     raise ScriptError(rval=CLIENT_INSTALL_ERROR)

        # Do not ask for time source
        # if options.conf_ntp:
        #     if not options.on_master and not options.unattended and not (
        #             options.ntp_servers or options.ntp_pool):
        #         options.ntp_servers, options.ntp_pool = \
        #             timeconf.get_time_source()

        cli_realm = ds.realm
        cli_realm_source = ds.realm_source
        logger.debug("will use discovered realm: %s", cli_realm)

        if options.realm_name and options.realm_name != cli_realm:
            logger.error(
                "The provided realm name [%s] does not match discovered "
                "one [%s]",
                options.realm_name, cli_realm)
            logger.debug("(%s: %s)", cli_realm, cli_realm_source)
            raise ScriptError(
                "The provided realm name [%s] does not match discovered "
                "one [%s]" % (options.realm_name, cli_realm),
                rval=CLIENT_INSTALL_ERROR)

        cli_basedn = ds.basedn
        cli_basedn_source = ds.basedn_source
        logger.debug("will use discovered basedn: %s", cli_basedn)
        # subject_base = DN(('O', cli_realm))

        logger.info("Client hostname: %s", hostname)
        logger.debug("Hostname source: %s", hostname_source)
        logger.info("Realm: %s", cli_realm)
        logger.debug("Realm source: %s", cli_realm_source)
        logger.info("DNS Domain: %s", cli_domain)
        logger.debug("DNS Domain source: %s", cli_domain_source)
        logger.info("IPA Server: %s", ', '.join(cli_server))
        logger.debug("IPA Server source: %s", cli_server_source)
        logger.info("BaseDN: %s", cli_basedn)
        logger.debug("BaseDN source: %s", cli_basedn_source)

        if not options.on_master:
            if options.ntp_servers:
                for server in options.ntp_servers:
                    logger.info("NTP server: %s", server)

            if options.ntp_pool:
                logger.info("NTP pool: %s", options.ntp_pool)

        # ipa-join would fail with IP address instead of a FQDN
        for srv in cli_server:
            try:
                socket.inet_pton(socket.AF_INET, srv)
                is_ipaddr = True
            except socket.error:
                try:
                    socket.inet_pton(socket.AF_INET6, srv)
                    is_ipaddr = True
                except socket.error:
                    is_ipaddr = False

            if is_ipaddr:
                logger.warning(
                    "It seems that you are using an IP address "
                    "instead of FQDN as an argument to --server. The "
                    "installation may fail.")
                break

        # logger.info()
        # if not options.unattended and not user_input(
        #     "Continue to configure the system with these values?", False):
        #     raise ScriptError(rval=CLIENT_INSTALL_ERROR)

    except ScriptError as e:
        module.fail_json(msg=str(e))

    #########################################################################

    # client._install

    # May not happen in here at this time
    # if not options.on_master:
    #     # Try removing old principals from the keytab
    #     purge_host_keytab(cli_realm)

    # Check if ipa client is already configured
    if is_client_configured():
        client_already_configured = True

        # Check that realm and domain match
        current_config = get_ipa_conf()
        if cli_domain != current_config.get('domain'):
            module.fail_json(msg="IPA client already installed "
                             "with a conflicting domain")
        if cli_realm != current_config.get('realm'):
            module.fail_json(msg="IPA client already installed "
                             "with a conflicting realm")
    else:
        client_already_configured = False

    # Done
    module.exit_json(changed=False,
                     servers=cli_server,
                     domain=cli_domain,
                     realm=cli_realm,
                     kdc=cli_kdc,
                     basedn=str(cli_basedn),
                     hostname=hostname,
                     client_domain=client_domain,
                     dnsok=dnsok,
                     sssd=options.sssd,
                     ntp_servers=options.ntp_servers,
                     ntp_pool=options.ntp_pool,
                     client_already_configured=client_already_configured,
                     ipa_python_version=IPA_PYTHON_VERSION,
                     nosssd_files=nosssd_files)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Based on ipa-client-install code
#
# Copyright (C) 2017-2022  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.0',
    'supported_by': 'community',
    'status': ['preview'],
}

DOCUMENTATION = '''
---
module: ipaserver_setup_dns
short_description: Setup DNS
description: Setup DNS
options:
  ip_addresses:
    description: List of Master Server IP Addresses
    type: list
    elements: str
    required: no
  domain:
    description: Primary DNS domain of the IPA deployment
    type: str
    required: yes
  realm:
    description: Kerberos realm name of the IPA deployment
    type: str
    required: yes
  hostname:
    description: Fully qualified name of this host
    type: str
    required: yes
  setup_dns:
    description: Configure bind with our zone
    type: bool
    required: yes
  setup_ca:
    description: Configure a dogtag CA
    type: bool
    required: yes
  zonemgr:
    description: DNS zone manager e-mail address. Defaults to hostmaster@DOMAIN
    type: str
    required: no
  forwarders:
    description: Add DNS forwarders
    type: list
    elements: str
    required: yes
  forward_policy:
    description: DNS forwarding policy for global forwarders
    type: str
    choices: ['first', 'only']
    default: 'first'
    required: no
  no_dnssec_validation:
    description: Disable DNSSEC validation
    type: bool
    default: no
    required: no
  dns_ip_addresses:
    description: The dns ip_addresses setting
    type: list
    elements: str
    required: yes
  dns_reverse_zones:
    description: The dns reverse_zones setting
    type: list
    elements: str
    required: yes
author:
    - Thomas Woerner (@t-woerner)
'''

EXAMPLES = '''
'''

RETURN = '''
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.ansible_ipa_server import (
    check_imports, AnsibleModuleLog, setup_logging, options, paths, dns,
    ansible_module_get_parsed_ip_addresses, sysrestore, api_Backend_ldap2,
    redirect_stdout, bindinstance
)


def main():
    ansible_module = AnsibleModule(
        argument_spec=dict(
            # basic
            ip_addresses=dict(required=False, type='list', elements='str',
                              default=[]),
            domain=dict(required=True, type='str'),
            realm=dict(required=True, type='str'),
            hostname=dict(required=True, type='str'),
            # server
            setup_dns=dict(required=True, type='bool'),
            setup_ca=dict(required=True, type='bool'),
            # dns
            zonemgr=dict(required=False, type='str'),
            forwarders=dict(required=True, type='list', elements='str'),
            forward_policy=dict(required=False, choices=['first', 'only'],
                                default='first'),
            no_dnssec_validation=dict(required=False, type='bool',
                                      default=False),
            # additional
            dns_ip_addresses=dict(required=True, type='list', elements='str'),
            dns_reverse_zones=dict(required=True, type='list', elements='str'),
        ),
    )

    ansible_module._ansible_debug = True
    check_imports(ansible_module)
    setup_logging()
    ansible_log = AnsibleModuleLog(ansible_module)

    # set values ############################################################

    # basic
    options.ip_addresses = ansible_module_get_parsed_ip_addresses(
        ansible_module)
    options.domain_name = ansible_module.params.get('domain')
    options.realm_name = ansible_module.params.get('realm')
    options.host_name = ansible_module.params.get('hostname')
    # server
    options.setup_dns = ansible_module.params.get('setup_dns')
    options.setup_ca = ansible_module.params.get('setup_ca')
    # dns
    options.zonemgr = ansible_module.params.get('zonemgr')
    options.forwarders = ansible_module.params.get('forwarders')
    options.forward_policy = ansible_module.params.get('forward_policy')
    options.no_dnssec_validation = ansible_module.params.get(
        'no_dnssec_validation')
    # additional
    dns.ip_addresses = ansible_module_get_parsed_ip_addresses(
        ansible_module, 'dns_ip_addresses')
    dns.reverse_zones = ansible_module.params.get('dns_reverse_zones')

    # init ##################################################################

    fstore = sysrestore.FileStore(paths.SYSRESTORE)

    api_Backend_ldap2(options.host_name, options.setup_ca, connect=True)

    # setup dns #############################################################

    with redirect_stdout(ansible_log):
        if options.setup_dns:
            dns.install(False, False, options)
        else:
            # Create a BIND instance
            bind = bindinstance.BindInstance(fstore)
            bind.set_output(ansible_log)
            bind.setup(options.host_name, options.ip_addresses,
                       options.realm_name,
                       options.domain_name, (), 'first', (),
                       zonemgr=options.zonemgr,
                       no_dnssec_validation=options.no_dnssec_validation)
            bind.create_file_with_system_records()

    # done ##################################################################

    ansible_module.exit_json(changed=True)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Based on ipa-client-install code
#
# Copyright (C) 2017-2022  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.0',
    'supported_by': 'community',
    'status': ['preview'],
}

DOCUMENTATION = '''
---
module: ipaserver_setup_ntp
short_description: Setup NTP
description: Setup NTP
options:
  ntp_servers:
    description: ntp servers to use
    type: list
    elements: str
    required: no
  ntp_pool:
    description: ntp server pool to use
    type: str
    required: no
author:
    - Thomas Woerner (@t-woerner)
'''

EXAMPLES = '''
'''

RETURN = '''
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.ansible_ipa_server import (
    check_imports, AnsibleModuleLog, setup_logging, options, sysrestore, paths,
    redirect_stdout, time_service, sync_time, ntpinstance, timeconf,
    getargspec
)


def main():
    ansible_module = AnsibleModule(
        argument_spec=dict(
            ntp_servers=dict(required=False, type='list', elements='str',
                             default=None),
            ntp_pool=dict(required=False, type='str', default=None),
        ),
    )

    ansible_module._ansible_debug = True
    check_imports(ansible_module)
    setup_logging()
    ansible_log = AnsibleModuleLog(ansible_module)

    # set values ############################################################

    options.ntp_servers = ansible_module.params.get('ntp_servers')
    options.ntp_pool = ansible_module.params.get('ntp_pool')

    # init ##########################################################

    fstore = sysrestore.FileStore(paths.SYSRESTORE)
    sstore = sysrestore.StateFile(paths.SYSRESTORE)

    # setup NTP #####################################################

    if time_service == "chronyd":
        # We have to sync time before certificate handling on master.
        # As chrony configuration is moved from client here, unconfiguration of
        # chrony will be handled here in uninstall() method as well by invoking
        # the ipa-server-install --uninstall
        ansible_module.log("Synchronizing time")

        # pylint: disable=deprecated-method
        argspec = getargspec(sync_time)
        # pylint: enable=deprecated-method
        if "options" not in argspec.args:
            synced_ntp = sync_time(options.ntp_servers, options.ntp_pool,
                                   fstore, sstore)
        else:
            synced_ntp = sync_time(options, fstore, sstore)
        if not synced_ntp:
            ansible_module.log(
                "Warning: IPA was unable to sync time with chrony!")
            ansible_module.log(
                "         Time synchronization is required for IPA "
                "to work correctly")
    else:
        # Configure ntpd
        timeconf.force_ntpd(sstore)
        ntp = ntpinstance.NTPInstance(fstore)
        ntp.set_output(ansible_log)
        with redirect_stdout(ansible_log):
            if not ntp.is_configured():
                ntp.create_instance()

    # done ##########################################################

    ansible_module.exit_json(changed=True)


if __name__ == '__main__':
    main()
